#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import logging

from logging import Logger
from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Set, Union

import pandas as pd
from ax.core.base_trial import BaseTrial
from ax.core.map_data import MapData, MapKeyInfo
from ax.core.map_metric import MapMetric
from ax.core.metric import Metric, MetricFetchE, MetricFetchResult
from ax.core.trial import Trial
from ax.metrics.curve import AbstractCurveMetric
from ax.utils.common.logger import get_logger
from ax.utils.common.result import Err, Ok
from pyre_extensions import assert_is_instance

logger: Logger = get_logger(__name__)

SMOOTHING_DEFAULT = 0.6  # Default in Tensorboard UI
RUN_METADATA_KEY = "tb_log_dir"

try:
    from tensorboard.backend.event_processing import (
        plugin_event_multiplexer as event_multiplexer,
    )
    from tensorboard.compat.proto import types_pb2

    logging.getLogger("tensorboard").setLevel(logging.CRITICAL)

    class TensorboardMetric(MapMetric):
        """A *new* `MapMetric` for getting Tensorboard metrics."""

        map_key_info: MapKeyInfo[float] = MapKeyInfo(key="step", default_value=0.0)

        def __init__(
            self,
            name: str,
            tag: str,
            lower_is_better: Optional[bool] = True,
            smoothing: float = SMOOTHING_DEFAULT,
            cumulative_best: bool = False,
        ) -> None:
            """
            Args:
                name: The name of the metric.
                tag: The name of the learning curve in the Tensorboard Scalars tab.
                lower_is_better: If True, lower curve values are considered better.
                smoothing: If > 0, apply exponential weighted mean to the curve. This
                    is the same postprocessing as the "smoothing" slider in the
                    Tensorboard UI.
                cumulative_best: If True, for each trial, apply cumulative best to
                    the curve (i.e., if lower is better, then we return a curve
                    representing the cumulative min of the raw curve).
            """
            super().__init__(name=name, lower_is_better=lower_is_better)

            self.smoothing = smoothing
            self.tag = tag
            self.cumulative_best = cumulative_best

        @classmethod
        def is_available_while_running(cls) -> bool:
            return True

        def bulk_fetch_trial_data(
            self, trial: BaseTrial, metrics: List[Metric], **kwargs: Any
        ) -> Dict[str, MetricFetchResult]:
            """Fetch multiple metrics data for one trial, using instance attributes
            of the metrics.

            Returns Dict of metric_name => Result
            Default behavior calls `fetch_trial_data` for each metric. Subclasses should
            override this to perform trial data computation for multiple metrics.
            """
            tb_metrics = [
                assert_is_instance(metric, TensorboardMetric) for metric in metrics
            ]

            trial = assert_is_instance(trial, Trial)
            if trial.arm is None:
                raise ValueError("Trial must have arm set.")

            arm_name = trial.arm.name

            try:
                mul = self._get_event_multiplexer_for_trial(trial=trial)
            except Exception as e:
                return {
                    metric.name: Err(
                        MetricFetchE(
                            message=f"Failed to get event multiplexer for {trial=}",
                            exception=e,
                        )
                    )
                    for metric in tb_metrics
                }

            scalar_dict = mul.PluginRunToTagToContent("scalars")
            if len(scalar_dict) == 0:
                return {
                    metric.name: Err(
                        MetricFetchE(
                            message=(
                                "No 'scalar' data found for trial in multiplexer "
                                f"{mul=}"
                            ),
                            exception=None,
                        )
                    )
                    for metric in tb_metrics
                }

            res = {}
            for metric in tb_metrics:
                try:
                    records = [
                        {
                            "trial_index": trial.index,
                            "arm_name": arm_name,
                            "metric_name": metric.name,
                            self.map_key_info.key: t.step,
                            "mean": (
                                t.tensor_proto.double_val[0]
                                if t.tensor_proto.double_val
                                else t.tensor_proto.float_val[0]
                            ),
                            "sem": float("nan"),
                        }
                        for run_name, tags in scalar_dict.items()
                        for tag in tags
                        if tag == metric.tag
                        for t in mul.Tensors(run_name, tag)
                    ]

                    # If records is empty something has gone wrong: either the tag is
                    # not present on the multiplexer or the content referenced is empty
                    if len(records) == 0:
                        if metric.tag not in [
                            j for sub in scalar_dict.values() for j in sub
                        ]:
                            raise KeyError(
                                f"Tag {metric.tag} not found on multiplexer {mul=}. "
                                "Did you specify this tag exactly as it appears in "
                                "the TensorBoard UI's Scalars tab?"
                            )
                        else:
                            raise ValueError(
                                f"Found tag {metric.tag}, but no data found for it. Is "
                                "the curve empty in the TensorBoard UI?"
                            )

                    df = (
                        pd.DataFrame(records)
                        # If a metric has multiple records for the same arm, metric, and
                        # step (sometimes caused by restarts, etc) take the mean
                        .groupby(["arm_name", "metric_name", self.map_key_info.key])
                        .mean()
                        .reset_index()
                    )

                    # Apply per-metric post-processing
                    # Apply cumulative "best" (min if lower_is_better)
                    if metric.cumulative_best:
                        if metric.lower_is_better:
                            df["mean"] = df["mean"].cummin()
                        else:
                            df["mean"] = df["mean"].cummax()

                    # Apply smoothing
                    if metric.smoothing > 0:
                        df["mean"] = df["mean"].ewm(alpha=metric.smoothing).mean()

                    # Accumulate successfully extracted timeseries
                    res[metric.name] = Ok(
                        MapData(
                            df=df,
                            map_key_infos=[self.map_key_info],
                        )
                    )

                except Exception as e:
                    res[metric.name] = Err(
                        MetricFetchE(
                            message=f"Failed to fetch data for {metric.name}",
                            exception=e,
                        )
                    )

            return res

        def fetch_trial_data(
            self, trial: BaseTrial, **kwargs: Any
        ) -> MetricFetchResult:
            """Fetch data for one trial."""

            return self.bulk_fetch_trial_data(trial=trial, metrics=[self], **kwargs)[
                self.name
            ]

        def _get_event_multiplexer_for_trial(
            self, trial: BaseTrial
        ) -> event_multiplexer.EventMultiplexer:
            """Get an event multiplexer with the logs for a given trial."""

            mul = event_multiplexer.EventMultiplexer(max_reload_threads=20)
            mul.AddRunsFromDirectory(trial.run_metadata[RUN_METADATA_KEY], None)
            mul.Reload()

            return mul

    class TensorboardCurveMetric(AbstractCurveMetric):
        """A `CurveMetric` for getting Tensorboard curves."""

        map_key_info: MapKeyInfo[float] = MapKeyInfo(key="steps", default_value=0.0)

        def get_curves_from_ids(
            self,
            ids: Iterable[Union[int, str]],
            names: Optional[Set[str]] = None,
        ) -> Dict[Union[int, str], Dict[str, pd.Series]]:
            """Get curve data from tensorboard logs.

            NOTE: If the ids are not simple paths/posix locations, subclass this metric
            and replace this method with an appropriate one that retrieves the log
            results.

            Args:
                ids: A list of string paths to tensorboard log directories.
                names: The names of the tags for which to fetch the curves.
                    If omitted, all tags are returned.

            Returns:
                A nested dictionary mapping ids (first level) and metric names (second
                level) to pandas Series of data.
            """
            return {idx: get_tb_from_posix(path=str(idx), tags=names) for idx in ids}

    def get_tb_from_posix(
        path: str,
        tags: Optional[Set[str]] = None,
    ) -> Dict[str, pd.Series]:
        r"""Get Tensorboard data from a posix path.

        Args:
            path: The posix path for the directory that contains the tensorboard logs.
            tags: The names of the tags for which to fetch the curves. If omitted,
                all tags are returned.
        Returns:
            A dictionary mapping tag names to pandas Series of data.
        """
        logger.debug(f"Reading TB logs from {path}.")
        mul = event_multiplexer.EventMultiplexer(max_reload_threads=20)
        mul.AddRunsFromDirectory(path, None)
        mul.Reload()
        scalar_dict = mul.PluginRunToTagToContent("scalars")

        raw_result = [
            {"tag": tag, "event": mul.Tensors(run, tag)}
            for run, run_dict in scalar_dict.items()
            for tag in run_dict
            if tags is None or tag in tags
        ]
        tb_run_data = {}
        for item in raw_result:
            latest_start_time = _get_latest_start_time(item["event"])
            steps = [e.step for e in item["event"] if e.wall_time >= latest_start_time]
            vals = [
                _get_event_value(e)
                for e in item["event"]
                if e.wall_time >= latest_start_time
            ]
            key = item["tag"]
            series = pd.Series(index=steps, data=vals).dropna()
            if key in tb_run_data:
                tb_run_data[key] = pd.concat(objs=[tb_run_data[key], series])
            else:
                tb_run_data[key] = series
        for key, series in tb_run_data.items():
            if any(series.index.duplicated()):
                # take average of repeated observations of the same "step"
                series = series.groupby(series.index).mean()
                logger.debug(
                    f"Found duplicate steps for tag {key}. "
                    "Removing duplicates by averaging."
                )
                tb_run_data[key] = series
        return tb_run_data

    # pyre-fixme[24]: Generic type `list` expects 1 type parameter, use
    #  `typing.List` to avoid runtime subscripting errors.
    def _get_latest_start_time(events: List) -> float:
        """In each directory, there may be previous training runs due to restarting
        training jobs.

        Args:
            events: A list of TensorEvents.

        Returns:
            The start time of the latest training run.
        """
        events.sort(key=lambda e: e.wall_time)
        start_time = events[0].wall_time
        for i in range(1, len(events)):
            # detect points in time where restarts occurred
            if events[i].step < events[i - 1].step:
                start_time = events[i].wall_time
        return start_time

    def _get_event_value(e: NamedTuple) -> float:
        r"""Helper function to check the dtype and then get the value
        stored in a TensorEvent."""
        tensor = e.tensor_proto  # pyre-ignore[16]
        if tensor.dtype == types_pb2.DT_FLOAT:
            return tensor.float_val[0]
        elif tensor.dtype == types_pb2.DT_DOUBLE:
            return tensor.double_val[0]
        elif tensor.dtype == types_pb2.DT_INT32:
            return tensor.int_val[0]
        else:
            raise ValueError(f"Tensorboard dtype {tensor.dtype} not supported.")

except ImportError:
    logger.warning(
        "tensorboard package not found. If you would like to use "
        "TensorboardCurveMetric, please install tensorboard."
    )
    pass



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import warnings

from collections import defaultdict, OrderedDict
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from logging import Logger
from typing import (
    DefaultDict,
    Dict,
    List,
    MutableMapping,
    Optional,
    Set,
    TYPE_CHECKING,
    Union,
)

import numpy as np
from ax.core.arm import Arm
from ax.core.base_trial import BaseTrial
from ax.core.data import Data
from ax.core.generator_run import ArmWeight, GeneratorRun, GeneratorRunType
from ax.core.trial import immutable_once_run
from ax.core.types import (
    TCandidateMetadata,
    TEvaluationOutcome,
    validate_evaluation_outcome,
)
from ax.exceptions.core import AxError, UnsupportedError, UserInputError
from ax.utils.common.base import SortableBase
from ax.utils.common.docutils import copy_doc
from ax.utils.common.equality import datetime_equals, equality_typechecker
from ax.utils.common.logger import _round_floats_for_logging, get_logger
from ax.utils.common.typeutils import checked_cast, not_none

logger: Logger = get_logger(__name__)


if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import core  # noqa F401

BATCH_TRIAL_RAW_DATA_FORMAT_ERROR_MESSAGE = (
    "Raw data must be a dict for batched trials."
)


class LifecycleStage(int, Enum):
    EXPLORATION = 0
    ITERATION = 1
    BAKEOFF = 2
    OFFLINE_OPTIMIZED = 3
    EXPLORATION_CONCURRENT = 4


@dataclass
class AbandonedArm(SortableBase):
    """Class storing metadata of arm that has been abandoned within
    a BatchTrial.
    """

    name: str
    time: datetime
    reason: Optional[str] = None

    @equality_typechecker
    def __eq__(self, other: AbandonedArm) -> bool:
        return (
            self.name == other.name
            and self.reason == other.reason
            and datetime_equals(self.time, other.time)
        )

    @property
    def _unique_id(self) -> str:
        return self.name


@dataclass
class GeneratorRunStruct(SortableBase):
    """Stores GeneratorRun object as well as the weight with which it was added."""

    generator_run: GeneratorRun
    weight: float

    @property
    def _unique_id(self) -> str:
        return self.generator_run._unique_id + ":" + str(self.weight)


class BatchTrial(BaseTrial):
    """Batched trial that has multiple attached arms, meant to be
    *deployed and evaluated together*, and possibly arm weights, which are
    a measure of how much of the total resources allocated to evaluating
    a batch should go towards evaluating the specific arm. For instance,
    for field experiments the weights could describe the fraction of the
    total experiment population assigned to the different treatment arms.
    Interpretation of the weights is defined in Runner.

    NOTE: A `BatchTrial` is not just a trial with many arms; it is a trial,
    for which it is important that the arms are evaluated simultaneously, e.g.
    in an A/B test where the evaluation results are subject to nonstationarity.
    For cases where multiple arms are evaluated separately and independently of
    each other, use multiple `Trial` objects with a single arm each.

    Args:
        experiment: Experiment, to which this trial is attached
        generator_run: GeneratorRun, associated with this trial. This can a
            also be set later through `add_arm` or `add_generator_run`, but a
            trial's associated generator run is immutable once set.
        generator_runs: GeneratorRuns, associated with this trial. This can a
            also be set later through `add_arm` or `add_generator_run`, but a
            trial's associated generator run is immutable once set.  This cannot
            be combined with the `generator_run` argument.
        trial_type: Type of this trial, if used in MultiTypeExperiment.
        optimize_for_power: Whether to optimize the weights of arms in this
            trial such that the experiment's power to detect effects of
            certain size is as high as possible. Refer to documentation of
            `BatchTrial.set_status_quo_and_optimize_power` for more detail.
        ttl_seconds: If specified, trials will be considered failed after
            this many seconds since the time the trial was ran, unless the
            trial is completed before then. Meant to be used to detect
            'dead' trials, for which the evaluation process might have
            crashed etc., and which should be considered failed after
            their 'time to live' has passed.
        index: If specified, the trial's index will be set accordingly.
            This should generally not be specified, as in the index will be
            automatically determined based on the number of existing trials.
            This is only used for the purpose of loading from storage.
        lifecycle_stage: The stage of the experiment lifecycle that this
            trial represents
    """

    def __init__(
        self,
        experiment: core.experiment.Experiment,
        generator_run: Optional[GeneratorRun] = None,
        generator_runs: Optional[List[GeneratorRun]] = None,
        trial_type: Optional[str] = None,
        optimize_for_power: Optional[bool] = False,
        ttl_seconds: Optional[int] = None,
        index: Optional[int] = None,
        lifecycle_stage: Optional[LifecycleStage] = None,
    ) -> None:
        super().__init__(
            experiment=experiment,
            trial_type=trial_type,
            ttl_seconds=ttl_seconds,
            index=index,
        )
        self._arms_by_name: Dict[str, Arm] = {}
        self._generator_run_structs: List[GeneratorRunStruct] = []
        self._abandoned_arms_metadata: Dict[str, AbandonedArm] = {}
        self._status_quo: Optional[Arm] = None
        self._status_quo_weight_override: Optional[float] = None
        if generator_run is not None:
            if generator_runs is not None:
                raise UnsupportedError(
                    "Cannot specify both `generator_run` and `generator_runs`."
                )
            self.add_generator_run(generator_run=generator_run)
        elif generator_runs is not None:
            for gr in generator_runs:
                self.add_generator_run(generator_run=gr)

        self.optimize_for_power = optimize_for_power
        status_quo = experiment.status_quo
        if optimize_for_power:
            if status_quo is None:
                raise ValueError(
                    "Can only optimize for power if experiment has a status quo."
                )
            self.set_status_quo_and_optimize_power(status_quo=status_quo)
        else:
            # Set the status quo for tracking purposes
            # It will not be included in arm_weights
            self._status_quo = status_quo

        # Trial status quos are stored in the DB as a generator run
        # with one arm; thus we need to store two `db_id` values
        # for this object instead of one
        self._status_quo_generator_run_db_id: Optional[int] = None
        self._status_quo_arm_db_id: Optional[int] = None
        self._lifecycle_stage = lifecycle_stage

    @property
    def experiment(self) -> core.experiment.Experiment:
        """The experiment this batch belongs to."""
        return self._experiment

    @property
    def index(self) -> int:
        """The index of this batch within the experiment's batch list."""
        return self._index

    @property
    def generator_run_structs(self) -> List[GeneratorRunStruct]:
        """List of generator run structs attached to this trial.

        Struct holds generator_run object and the weight with which it was added.
        """
        return self._generator_run_structs

    @property
    def arm_weights(self) -> MutableMapping[Arm, float]:
        """The set of arms and associated weights for the trial.

        These are constructed by merging the arms and weights from
        each generator run that is attached to the trial.
        """
        arm_weights = OrderedDict()
        if len(self._generator_run_structs) == 0 and self.status_quo is None:
            return arm_weights
        for struct in self._generator_run_structs:
            multiplier = struct.weight
            for arm, weight in struct.generator_run.arm_weights.items():
                scaled_weight = weight * multiplier
                if arm in arm_weights:
                    arm_weights[arm] += scaled_weight
                else:
                    arm_weights[arm] = scaled_weight
        if self.status_quo is not None and self._status_quo_weight_override is not None:
            # If override is specified, this is the weight the status quo gets,
            # regardless of whether it appeared in any generator runs.
            # If no override is specified, status quo does not appear in arm_weights.
            arm_weights[self.status_quo] = self._status_quo_weight_override
        return arm_weights

    @property
    def lifecycle_stage(self) -> Optional[LifecycleStage]:
        return self._lifecycle_stage

    @arm_weights.setter
    def arm_weights(self, arm_weights: MutableMapping[Arm, float]) -> None:
        raise NotImplementedError("Use `trial.add_arms_and_weights`")

    @immutable_once_run
    def add_arm(self, arm: Arm, weight: float = 1.0) -> BatchTrial:
        """Add a arm to the trial.

        Args:
            arm: The arm to be added.
            weight: The weight with which this arm should be added.

        Returns:
            The trial instance.
        """
        return self.add_arms_and_weights(arms=[arm], weights=[weight])

    @immutable_once_run
    def add_arms_and_weights(
        self,
        arms: List[Arm],
        weights: Optional[List[float]] = None,
        multiplier: float = 1.0,
    ) -> BatchTrial:
        """Add arms and weights to the trial.

        Args:
            arms: The arms to be added.
            weights: The weights associated with the arms.
            multiplier: The multiplier applied to input weights before merging with
                the current set of arms and weights.

        Returns:
            The trial instance.
        """
        return self.add_generator_run(
            generator_run=GeneratorRun(
                arms=arms, weights=weights, type=GeneratorRunType.MANUAL.name
            ),
            multiplier=multiplier,
        )

    @immutable_once_run
    def add_generator_run(
        self, generator_run: GeneratorRun, multiplier: float = 1.0
    ) -> BatchTrial:
        """Add a generator run to the trial.

        The arms and weights from the generator run will be merged with
        the existing arms and weights on the trial, and the generator run
        object will be linked to the trial for tracking.

        Args:
            generator_run: The generator run to be added.
            multiplier: The multiplier applied to input weights before merging with
                the current set of arms and weights.

        Returns:
            The trial instance.
        """
        # First validate generator run arms
        for arm in generator_run.arms:
            self.experiment.search_space.check_types(arm.parameters, raise_error=True)

        # Clone arms to avoid mutating existing state
        generator_run._arm_weight_table = OrderedDict(
            {
                arm_sig: ArmWeight(arm_weight.arm.clone(), arm_weight.weight)
                for arm_sig, arm_weight in generator_run._arm_weight_table.items()
            }
        )

        # Add names to arms
        # For those not yet added to this experiment, create a new name
        # Else, use the name of the existing arm
        for arm in generator_run.arms:
            self._check_existing_and_name_arm(arm)

        self._generator_run_structs.append(
            GeneratorRunStruct(generator_run=generator_run, weight=multiplier)
        )
        generator_run.index = len(self._generator_run_structs) - 1

        if self.status_quo is not None and self.optimize_for_power:
            self.set_status_quo_and_optimize_power(status_quo=not_none(self.status_quo))

        if generator_run._generation_step_index is not None:
            self._set_generation_step_index(
                generation_step_index=generator_run._generation_step_index
            )
        self._refresh_arms_by_name()
        return self

    @property
    def status_quo(self) -> Optional[Arm]:
        """The control arm for this batch."""
        return self._status_quo

    @status_quo.setter
    def status_quo(self, status_quo: Optional[Arm]) -> None:
        raise NotImplementedError(
            "Use `set_status_quo_with_weight` or "
            "`set_status_quo_and_optimize_power` "
            "to set the status quo arm."
        )

    def unset_status_quo(self) -> None:
        """Set the status quo to None."""
        self._status_quo = None
        self._refresh_arms_by_name()

    @immutable_once_run
    def set_status_quo_with_weight(
        self, status_quo: Arm, weight: Optional[float]
    ) -> BatchTrial:
        """Sets status quo arm with given weight. This weight *overrides* any
        weight the status quo has from generator runs attached to this batch.
        Thus, this function is not the same as using add_arm, which will
        result in the weight being additive over all generator runs.
        """
        # Assign a name to this arm if none exists
        if weight is not None and weight <= 0.0:
            raise ValueError("Status quo weight must be positive.")

        if status_quo is not None:
            self.experiment.search_space.check_types(
                status_quo.parameters, raise_error=True
            )
            self.experiment._name_and_store_arm_if_not_exists(
                arm=status_quo, proposed_name="status_quo_" + str(self.index)
            )
        self._status_quo = status_quo.clone() if status_quo is not None else None
        self._status_quo_weight_override = weight
        self._refresh_arms_by_name()
        return self

    @immutable_once_run
    def set_status_quo_and_optimize_power(self, status_quo: Arm) -> BatchTrial:
        """Adds a status quo arm to the batch and optimizes for power.

        NOTE: this optimization based on the arms that are currently attached
        to the batch. If you add more arms later, you should re-run this function.
        If you want the optimization to happen automatically,
        set batch.optimize_for_power = True.

        This function will maximize power across the multiple pair-wise
        comparisons of existing arms against the status_quo.

        Specifically, this function assigns sqrt(sum_weights) weight to the
        status quo, where sum_weights is the sum of the weights of the existing
        arms, excluding the status quo. This will be optimal in terms of
        statistical power in the case where:
            1) status quo is the only arm to compare against,
            2) all other arms are of equal interest.
        """

        status_quo_is_only_arm = len(self.arms) == 1 and self._status_quo is not None
        if len(self.arms) == 0 or status_quo_is_only_arm:
            # If status quo is the only arm, just set its weight to 1
            # Can't use logic below, because it will choose 0
            self.set_status_quo_with_weight(status_quo=status_quo, weight=1)
            return self

        # arm_weights should always have at least one arm now
        arm_weights = not_none(self.arm_weights)
        sum_weights = sum(w for arm, w in arm_weights.items() if arm != status_quo)
        optimal_status_quo_weight_override = np.sqrt(sum_weights)
        self.set_status_quo_with_weight(
            status_quo=status_quo, weight=optimal_status_quo_weight_override
        )
        return self

    @property
    def arms(self) -> List[Arm]:
        """All arms contained in the trial."""
        arm_weights = self.arm_weights
        return [] if arm_weights is None else list(arm_weights.keys())

    @property
    def weights(self) -> List[float]:
        """Weights corresponding to arms contained in the trial."""
        arm_weights = self.arm_weights
        return [] if arm_weights is None else list(arm_weights.values())

    @property
    def arms_by_name(self) -> Dict[str, Arm]:
        """Map from arm name to object for all arms in trial."""
        return self._arms_by_name

    def _refresh_arms_by_name(self) -> None:
        self._arms_by_name = {}
        for arm in self.arms:
            if not arm.has_name:
                raise ValueError("Arms attached to a trial must have a name.")
            self._arms_by_name[arm.name] = arm

    @property
    def abandoned_arms(self) -> List[Arm]:
        """List of arms that have been abandoned within this trial."""
        return [
            self.arms_by_name[arm.name]
            for arm in self._abandoned_arms_metadata.values()
        ]

    @property
    def abandoned_arm_names(self) -> Set[str]:
        """Set of names of arms that have been abandoned within this trial."""
        return set(self._abandoned_arms_metadata.keys())

    @property
    def in_design_arms(self) -> List[Arm]:
        return [
            arm
            for arm in self.arms
            if self.experiment.search_space.check_membership(arm.parameters)
        ]

    # pyre-ignore[6]: T77111662.
    @copy_doc(BaseTrial.generator_runs)
    @property
    def generator_runs(self) -> List[GeneratorRun]:
        return [grs.generator_run for grs in self.generator_run_structs]

    @property
    def abandoned_arms_metadata(self) -> List[AbandonedArm]:
        return list(self._abandoned_arms_metadata.values())

    @property
    def is_factorial(self) -> bool:
        """Return true if the trial's arms are a factorial design with
        no linked factors.
        """
        # To match the model behavior, this should probably actually be pulled
        # from exp.parameters. However, that seems rather ugly when this function
        # intuitively should just depend on the arms.
        sufficient_factors = all(len(arm.parameters or []) >= 2 for arm in self.arms)
        if not sufficient_factors:
            return False
        param_levels: DefaultDict[str, Dict[Union[str, float], int]] = defaultdict(dict)
        for arm in self.arms:
            for param_name, param_value in arm.parameters.items():
                param_levels[param_name][not_none(param_value)] = 1
        param_cardinality = 1
        for param_values in param_levels.values():
            param_cardinality *= len(param_values)
        return len(self.arms) == param_cardinality

    def run(self) -> BatchTrial:
        return checked_cast(BatchTrial, super().run())

    def normalized_arm_weights(
        self, total: float = 1, trunc_digits: Optional[int] = None
    ) -> MutableMapping[Arm, float]:
        """Returns arms with a new set of weights normalized
        to the given total.

        This method is useful for many runners where we need to normalize weights
        to a certain total without mutating the weights attached to a trial.

        Args:
            total: The total weight to which to normalize.
                Default is 1, in which case arm weights
                can be interpreted as probabilities.
            trunc_digits: The number of digits to keep. If the
                resulting total weight is not equal to `total`, re-allocate
                weight in such a way to maintain relative weights as best as
                possible.

        Returns:
            Mapping from arms to the new set of weights.

        """
        weights = np.array(self.weights)
        if trunc_digits is not None:
            atomic_weight = 10**-trunc_digits
            # pyre-fixme[16]: `float` has no attribute `astype`.
            int_weights = (
                (total / atomic_weight) * (weights / np.sum(weights))
            ).astype(int)
            n_leftover = int(total / atomic_weight) - np.sum(int_weights)
            int_weights[:n_leftover] += 1
            weights = int_weights * atomic_weight
        else:
            weights = weights * (total / np.sum(weights))
        return OrderedDict(zip(self.arms, weights))

    def mark_arm_abandoned(
        self, arm_name: str, reason: Optional[str] = None
    ) -> BatchTrial:
        """Mark a arm abandoned.

        Usually done after deployment when one arm causes issues but
        user wants to continue running other arms in the batch.

        NOTE: Abandoned arms are considered to be 'pending points' in
        experiment after their abandonment to avoid Ax models suggesting
        the same arm again as a new candidate. Abandoned arms are also
        excluded from model training data unless ``fit_abandoned``
        option is specified to model bridge.

        Args:
            arm_name: The name of the arm to abandon.
            reason: The reason for abandoning the arm.

        Returns:
            The batch instance.
        """
        if arm_name not in self.arms_by_name:
            raise ValueError("Arm must be contained in batch.")

        abandoned_arm = AbandonedArm(name=arm_name, time=datetime.now(), reason=reason)
        self._abandoned_arms_metadata[arm_name] = abandoned_arm
        return self

    def clone(self) -> BatchTrial:
        """Clone the trial and attach it to the current experiment."""
        warnings.warn(
            "clone() method is getting deprecated. Please use clone_to() instead.",
            DeprecationWarning,
            stacklevel=3,
        )
        return self.clone_to(include_sq=False)

    def clone_to(
        self,
        experiment: Optional[core.experiment.Experiment] = None,
        include_sq: bool = True,
    ) -> BatchTrial:
        """Clone the trial and attach it to a specified experiment.
        If None provided, attach it to the current experiment.

        Args:
            experiment: The experiment to which the cloned trial will belong.
                If unspecified, uses the current experiment.
            include_sq: Whether to include status quo in the cloned trial.

        Returns:
            A new instance of the trial.
        """
        use_old_experiment = experiment is None
        experiment = self._experiment if experiment is None else experiment
        new_trial = experiment.new_batch_trial(
            trial_type=self._trial_type, ttl_seconds=self._ttl_seconds
        )
        for struct in self._generator_run_structs:
            if use_old_experiment:
                # don't clone gen run in case we are attaching cloned trial to
                # the same experiment
                new_trial.add_generator_run(struct.generator_run, struct.weight)
            else:
                new_trial.add_generator_run(struct.generator_run.clone(), struct.weight)

        if (self._status_quo is not None) and include_sq:
            sq_weight = self._status_quo_weight_override
            new_trial.set_status_quo_with_weight(
                self._status_quo.clone(),
                weight=sq_weight,
            )
        self._update_trial_attrs_on_clone(new_trial=new_trial)
        return new_trial

    def attach_batch_trial_data(
        self,
        raw_data: Dict[str, TEvaluationOutcome],
        sample_sizes: Optional[Dict[str, int]] = None,
        metadata: Optional[Dict[str, Union[str, int]]] = None,
    ) -> None:
        """Attaches data to the trial

        Args:
            raw_data: Map from arm name to metric outcomes.
            sample_sizes: Dict from arm name to sample size.
            metadata: Additional metadata to track about this run.
                importantly the start_date and end_date
            complete_trial: Whether to mark trial as complete after
                attaching data. Defaults to False.
        """
        # Validate type of raw_data
        if not isinstance(raw_data, dict):
            raise ValueError(BATCH_TRIAL_RAW_DATA_FORMAT_ERROR_MESSAGE)

        for key, value in raw_data.items():
            if not isinstance(key, str):
                raise ValueError(BATCH_TRIAL_RAW_DATA_FORMAT_ERROR_MESSAGE)

            try:
                validate_evaluation_outcome(outcome=value)
            except TypeError:
                raise ValueError(BATCH_TRIAL_RAW_DATA_FORMAT_ERROR_MESSAGE)

        # Format the data to save.
        not_trial_arm_names = set(raw_data.keys()) - set(self.arms_by_name.keys())
        if not_trial_arm_names:
            raise UserInputError(
                f"Arms {not_trial_arm_names} are not part of trial #{self.index}."
            )

        evaluations, data = self._make_evaluations_and_data(
            raw_data=raw_data, metadata=metadata, sample_sizes=sample_sizes
        )
        self._validate_batch_trial_data(data=data)

        self._run_metadata = self._run_metadata if metadata is None else metadata
        self.experiment.attach_data(data)

        data_for_logging = _round_floats_for_logging(item=evaluations)

        logger.info(
            f"Updated trial {self.index} with data: "
            f"{_round_floats_for_logging(item=data_for_logging)}."
        )

    def __repr__(self) -> str:
        return (
            "BatchTrial("
            f"experiment_name='{self._experiment._name}', "
            f"index={self._index}, "
            f"status={self._status})"
        )

    def _get_candidate_metadata_from_all_generator_runs(
        self,
    ) -> Dict[str, TCandidateMetadata]:
        """Retrieves combined candidate metadata from all generator runs on this
        batch trial in the form of { arm name -> candidate metadata} mapping.

        NOTE: this does not handle the case of the same arm appearing in multiple
        generator runs in the same trial: metadata from only one of the generator
        runs containing the arm will be retrieved.
        """
        cand_metadata = {}
        for gr_struct in self._generator_run_structs:
            gr = gr_struct.generator_run
            if gr.candidate_metadata_by_arm_signature:
                gr_cand_metadata = gr.candidate_metadata_by_arm_signature
                warn = False
                for arm in gr.arms:
                    if arm.name in cand_metadata:
                        warn = True
                    if gr_cand_metadata:
                        # Reformat the mapping to be by arm name, since arm signature
                        # is not stored in Ax data.
                        cand_metadata[arm.name] = gr_cand_metadata.get(arm.signature)
                if warn:
                    logger.warning(
                        "The same arm appears in multiple generator runs in batch "
                        f"{self.index}. Candidate metadata will only contain metadata "
                        "for one of those generator runs, and the candidate metadata "
                        "for the arm from another generator run will not be propagated."
                    )
        return cand_metadata

    def _get_candidate_metadata(self, arm_name: str) -> TCandidateMetadata:
        """Retrieves candidate metadata for a specific arm."""
        try:
            arm = self.arms_by_name[arm_name]
        except KeyError:
            raise ValueError(
                f"Arm by name {arm_name} is not part of trial #{self.index}."
            )
        for gr_struct in self._generator_run_structs:
            gr = gr_struct.generator_run
            if gr and gr.candidate_metadata_by_arm_signature and arm in gr.arms:
                return not_none(gr.candidate_metadata_by_arm_signature).get(
                    arm.signature
                )
        return None

    def _validate_batch_trial_data(self, data: Data) -> None:
        """Utility function to validate batch data before further processing."""
        if (
            self.status_quo
            and not_none(self.status_quo).name in self.arms_by_name
            and not_none(self.status_quo).name not in data.df["arm_name"].values
        ):
            raise AxError(
                f"Trial #{self.index} was completed with data that did "
                "not contain status quo observations, but the trial has "
                "status quo set and therefore data for it is required."
            )

        for metric_name in data.df["metric_name"].values:
            if metric_name not in self.experiment.metrics:
                logger.info(
                    f"Data was logged for metric {metric_name} that was not yet "
                    "tracked on the experiment. Please specify `tracking_metric_"
                    "names` argument in AxClient.create_experiment to add tracking "
                    "metrics to the experiment. Without those, all data users "
                    "specify is still attached to the experiment, but will not be "
                    "fetched in `experiment.fetch_data()`, but you can still use "
                    "`experiment.lookup_data_for_trial` to get all attached data."
                )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numbers

from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd

import plotly.graph_objs as go
from ax.analysis.helpers.color_helpers import rgba

from ax.analysis.helpers.constants import CI_OPACITY, COLORS, DECIMALS, Z

from ax.analysis.helpers.plot_helpers import _format_CI, _format_dict

from ax.core.types import TParameterization

from ax.utils.stats.statstools import relativize

# disable false positive "SettingWithCopyWarning"
pd.options.mode.chained_assignment = None


def relativize_dataframe(df: pd.DataFrame, status_quo_name: str) -> pd.DataFrame:
    """
    Relativizes the dataframe with respect to "status_quo_name". Assumes as a
    precondition that for each metric in the dataframe, there is a row with
    arm_name == status_quo_name to relativize against.

    Args:
        df: dataframe with the following columns
            {
            "arm_name": name of the arm in the cross validation result
            "metric_name": name of the observed/predicted metric
            "x": value of the observation for the metric for this arm
            "x_se": standard error of the observation for the metric of this arm
            "y": value predicted for the metric for this arm
            "y_se": standard error of predicted metric for this arm
            }
        status_quo_name: name of the status quo arm in the dataframe to use
            for relativization.
    Returns:
        A dataframe containing the same rows as df, with the observation and predicted
        data values relativized with respect to the status quo.
        An additional column "rel" is added to indicate whether the data is relativized.

    """
    metrics = df["metric_name"].unique()

    def _relativize_filtered_dataframe(
        df: pd.DataFrame, metric_name: str, status_quo_name: str
    ) -> pd.DataFrame:
        df = df.loc[df["metric_name"] == metric_name]
        status_quo_row = df.loc[df["arm_name"] == status_quo_name]

        mean_c = status_quo_row["y"].iloc[0]
        sem_c = status_quo_row["y_se"].iloc[0]
        y_rel, y_se_rel = relativize(
            means_t=df["y"].tolist(),
            sems_t=df["y_se"].tolist(),
            mean_c=mean_c,
            sem_c=sem_c,
            as_percent=True,
        )
        df["y"] = y_rel
        df["y_se"] = y_se_rel

        mean_c = status_quo_row["x"].iloc[0]
        sem_c = status_quo_row["x_se"].iloc[0]
        x_rel, x_se_rel = relativize(
            means_t=df["x"].tolist(),
            sems_t=df["x_se"].tolist(),
            mean_c=mean_c,
            sem_c=sem_c,
            as_percent=True,
        )
        df["x"] = x_rel
        df["x_se"] = x_se_rel
        df["rel"] = True
        return df

    return pd.concat(
        [
            _relativize_filtered_dataframe(
                df=df, metric_name=metric, status_quo_name=status_quo_name
            )
            for metric in metrics
        ]
    )


def extract_mean_and_error_from_df(
    df: pd.DataFrame,
) -> Tuple[List[float], List[float], List[float], List[float]]:
    """Extract mean and error from dataframe.

    Args:
        df: dataframe containing the scatter plot data
    Returns:
        x: list of x values
        x_se: list of x standard error
        y: list of y values
        y_se: list of y standard error
    """
    x = df["x"]
    x_se = df["x_se"]
    y = df["y"]
    y_se = df["y_se"]

    return (x, x_se, y, y_se)


def make_label(
    arm_name: str,
    x_axis_values: Optional[Tuple[str, float, float]],
    y_axis_values: Tuple[str, float, float],
    param_blob: TParameterization,
    rel: bool,
) -> str:
    """Make label for scatter plot.

    Args:
        arm_name: Name of arm
        x_axis_values: Optional Tuple of
            x_name: Name of x variable
            x_val: Value of x variable
            x_se: Standard error of x variable
        y_axis_values: Tuple of
            y_name: Name of y variable
            y_val: Value of y variable
            y_se: Standard error of y variable
        param_blob: Parameterization of arm
        rel: whether the data is relativized as a %

    Returns:
        Label for scatter plot.
    """
    heading = f"<b>Arm {arm_name}</b><br>"
    x_lab = ""
    if x_axis_values is not None:
        x_name, x_val, x_se = x_axis_values
        x_lab = "{name}: {estimate}{perc} {ci}<br>".format(
            name=x_name,
            estimate=(
                round(x_val, DECIMALS) if isinstance(x_val, numbers.Number) else x_val
            ),
            ci=_format_CI(estimate=x_val, sd=x_se),
            perc="%" if rel else "",
        )

    y_name, y_val, y_se = y_axis_values
    y_lab = "{name}: {estimate}{perc} {ci}<br>".format(
        name=y_name,
        estimate=(
            round(y_val, DECIMALS) if isinstance(y_val, numbers.Number) else y_val
        ),
        ci=_format_CI(estimate=y_val, sd=y_se),
        perc="%" if rel else "",
    )

    parameterization = _format_dict(param_blob, "Parameterization")

    return "{arm_name}<br>{xlab}{ylab}{param_blob}".format(
        arm_name=heading,
        xlab=x_lab,
        ylab=y_lab,
        param_blob=parameterization,
    )


def error_dot_plot_trace_from_df(
    df: pd.DataFrame,
    show_CI: bool = True,
    visible: bool = True,
) -> Dict[str, Any]:
    """Creates trace for dot plot with confidence intervals.
    Categorizes by arm name.

    Args:
        df: dataframe containing the scatter plot data
        show_CI: if True, plot confidence intervals.
        visible: if True, trace will be visible in figure
    """

    _, _, y, y_se = extract_mean_and_error_from_df(df)

    labels = []

    metric_name = df["metric_name"].iloc[0]

    for _, row in df.iterrows():
        labels.append(
            make_label(
                arm_name=row["arm_name"],
                x_axis_values=(None),
                y_axis_values=(
                    metric_name,
                    row["y"],
                    row["y_se"],
                ),
                param_blob=row["arm_parameters"],
                rel=(False if "rel" not in row else row["rel"]),
            )
        )

    trace = go.Scatter(
        x=df["arm_name"],
        y=y,
        marker={"color": rgba(COLORS.STEELBLUE.value)},
        mode="markers",
        name="In-sample",
        text=labels,
        hoverinfo="text",
    )

    if show_CI:
        if y_se is not None:
            trace.update(
                error_y={
                    "type": "data",
                    "array": np.multiply(y_se, Z),
                    "color": rgba(COLORS.STEELBLUE.value, CI_OPACITY),
                }
            )

    trace.update(visible=visible)
    trace.update(showlegend=False)
    return trace


def error_scatter_trace_from_df(
    df: pd.DataFrame,
    show_CI: bool = True,
    visible: bool = True,
    y_axis_label: Optional[str] = None,
    x_axis_label: Optional[str] = None,
) -> Dict[str, Any]:
    """Plot scatterplot with error bars.

    Args:
        df: dataframe containing the scatter plot data
        show_CI: if True, plot confidence intervals.
        visible: if True, trace will be visible in figure
        y_axis_label: custom label to use for y axis.
            If None, use metric name from `y_axis_var`.
        x_axis_label: custom label to use for x axis.
            If None, use metric name from `x_axis_var` if that is not None.
    """

    x, x_se, y, y_se = extract_mean_and_error_from_df(df)

    labels = []

    metric_name = df["metric_name"].iloc[0]

    for _, row in df.iterrows():
        labels.append(
            make_label(
                arm_name=row["arm_name"],
                x_axis_values=(
                    metric_name if x_axis_label is None else x_axis_label,
                    row["x"],
                    row["x_se"],
                ),
                y_axis_values=(
                    (metric_name if y_axis_label is None else y_axis_label),
                    row["y"],
                    row["y_se"],
                ),
                param_blob=row["arm_parameters"],
                rel=(False if "rel" not in row else row["rel"]),
            )
        )

    trace = go.Scatter(
        x=x,
        y=y,
        marker={"color": rgba(COLORS.STEELBLUE.value)},
        mode="markers",
        name="In-sample",
        text=labels,
        hoverinfo="text",
    )

    if show_CI:
        if x_se is not None:
            trace.update(
                error_x={
                    "type": "data",
                    "array": np.multiply(x_se, Z),
                    "color": rgba(COLORS.STEELBLUE.value, CI_OPACITY),
                }
            )
        if y_se is not None:
            trace.update(
                error_y={
                    "type": "data",
                    "array": np.multiply(y_se, Z),
                    "color": rgba(COLORS.STEELBLUE.value, CI_OPACITY),
                }
            )

    trace.update(visible=visible)
    trace.update(showlegend=True)
    return trace



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, Dict, Optional

import numpy as np
from ax.benchmark.benchmark_method import BenchmarkMethod
from ax.benchmark.benchmark_problem import (
    BenchmarkProblem,
    MultiObjectiveBenchmarkProblem,
    SingleObjectiveBenchmarkProblem,
)
from ax.benchmark.benchmark_result import AggregatedBenchmarkResult, BenchmarkResult
from ax.benchmark.problems.surrogate import (
    MOOSurrogateBenchmarkProblem,
    SOOSurrogateBenchmarkProblem,
)
from ax.core.experiment import Experiment
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.modelbridge.generation_strategy import GenerationStep, GenerationStrategy
from ax.modelbridge.registry import Models
from ax.modelbridge.torch import TorchModelBridge
from ax.models.torch.botorch_modular.model import BoTorchModel
from ax.models.torch.botorch_modular.surrogate import Surrogate
from ax.service.scheduler import SchedulerOptions
from ax.utils.common.constants import Keys
from ax.utils.common.typeutils import checked_cast
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_branin_experiment_with_multi_objective,
)
from botorch.acquisition.monte_carlo import qNoisyExpectedImprovement
from botorch.models.gp_regression import SingleTaskGP
from botorch.test_functions.multi_objective import BraninCurrin, ConstrainedBraninCurrin
from botorch.test_functions.synthetic import Branin


def get_benchmark_problem() -> BenchmarkProblem:
    return BenchmarkProblem.from_botorch(
        test_problem_class=Branin,
        test_problem_kwargs={},
        lower_is_better=True,
        num_trials=4,
    )


def get_single_objective_benchmark_problem(
    observe_noise_sd: bool = False,
    num_trials: int = 4,
    test_problem_kwargs: Optional[Dict[str, Any]] = None,
) -> SingleObjectiveBenchmarkProblem:
    return SingleObjectiveBenchmarkProblem.from_botorch_synthetic(
        test_problem_class=Branin,
        test_problem_kwargs=test_problem_kwargs or {},
        lower_is_better=True,
        num_trials=num_trials,
        observe_noise_sd=observe_noise_sd,
    )


def get_multi_objective_benchmark_problem(
    observe_noise_sd: bool = False, num_trials: int = 4
) -> MultiObjectiveBenchmarkProblem:
    return MultiObjectiveBenchmarkProblem.from_botorch_multi_objective(
        test_problem_class=BraninCurrin,
        test_problem_kwargs={},
        num_trials=num_trials,
        observe_noise_sd=observe_noise_sd,
    )


def get_constrained_multi_objective_benchmark_problem(
    observe_noise_sd: bool = False, num_trials: int = 4
) -> MultiObjectiveBenchmarkProblem:
    return MultiObjectiveBenchmarkProblem.from_botorch_multi_objective(
        test_problem_class=ConstrainedBraninCurrin,
        test_problem_kwargs={},
        num_trials=num_trials,
        observe_noise_sd=observe_noise_sd,
    )


def get_sobol_benchmark_method() -> BenchmarkMethod:
    return BenchmarkMethod(
        name="SOBOL",
        generation_strategy=GenerationStrategy(
            steps=[GenerationStep(model=Models.SOBOL, num_trials=-1)],
            name="SOBOL",
        ),
        scheduler_options=SchedulerOptions(
            total_trials=4, init_seconds_between_polls=0
        ),
    )


def get_soo_surrogate() -> SOOSurrogateBenchmarkProblem:
    experiment = get_branin_experiment(with_completed_trial=True)
    surrogate = TorchModelBridge(
        experiment=experiment,
        search_space=experiment.search_space,
        model=BoTorchModel(surrogate=Surrogate(botorch_model_class=SingleTaskGP)),
        data=experiment.lookup_data(),
        transforms=[],
    )
    return SOOSurrogateBenchmarkProblem(
        name="test",
        search_space=experiment.search_space,
        optimization_config=checked_cast(
            OptimizationConfig, experiment.optimization_config
        ),
        num_trials=6,
        outcome_names=["branin"],
        observe_noise_stds=True,
        get_surrogate_and_datasets=lambda: (surrogate, []),
        optimal_value=0.0,
    )


def get_moo_surrogate() -> MOOSurrogateBenchmarkProblem:
    experiment = get_branin_experiment_with_multi_objective(with_completed_trial=True)
    surrogate = TorchModelBridge(
        experiment=experiment,
        search_space=experiment.search_space,
        model=BoTorchModel(surrogate=Surrogate(botorch_model_class=SingleTaskGP)),
        data=experiment.lookup_data(),
        transforms=[],
    )
    return MOOSurrogateBenchmarkProblem(
        name="test",
        search_space=experiment.search_space,
        optimization_config=checked_cast(
            MultiObjectiveOptimizationConfig, experiment.optimization_config
        ),
        num_trials=10,
        outcome_names=["branin_a", "branin_b"],
        observe_noise_stds=True,
        get_surrogate_and_datasets=lambda: (surrogate, []),
        maximum_hypervolume=1.0,
        reference_point=[],
    )


def get_sobol_gpei_benchmark_method() -> BenchmarkMethod:
    return BenchmarkMethod(
        name="MBO_SOBOL_GPEI",
        generation_strategy=GenerationStrategy(
            name="Modular::Sobol+GPEI",
            steps=[
                GenerationStep(
                    model=Models.SOBOL,
                    num_trials=3,
                    model_kwargs={"fit_tracking_metrics": False},
                    min_trials_observed=3,
                ),
                GenerationStep(
                    model=Models.BOTORCH_MODULAR,
                    num_trials=-1,
                    model_kwargs={
                        "surrogate": Surrogate(SingleTaskGP),
                        # TODO: tests should better reflect defaults and not
                        # re-implement this logic.
                        "botorch_acqf_class": qNoisyExpectedImprovement,
                        "model_kwargs": {"fit_tracking_metrics": False},
                    },
                    model_gen_kwargs={
                        "model_gen_options": {
                            Keys.OPTIMIZER_KWARGS: {
                                "num_restarts": 50,
                                "raw_samples": 1024,
                            },
                            Keys.ACQF_KWARGS: {
                                "prune_baseline": True,
                            },
                        }
                    },
                ),
            ],
        ),
        scheduler_options=SchedulerOptions(
            total_trials=4, init_seconds_between_polls=0
        ),
    )


def get_benchmark_result() -> BenchmarkResult:
    problem = get_single_objective_benchmark_problem()

    return BenchmarkResult(
        name="test_benchmarking_result",
        seed=0,
        experiment=Experiment(
            name="test_benchmarking_experiment",
            search_space=problem.search_space,
            optimization_config=problem.optimization_config,
            runner=problem.runner,
            is_test=True,
        ),
        optimization_trace=np.array([3, 2, 1, 0.1]),
        score_trace=np.array([3, 2, 1, 0.1]),
        fit_time=0.1,
        gen_time=0.2,
    )


def get_aggregated_benchmark_result() -> AggregatedBenchmarkResult:
    result = get_benchmark_result()
    return AggregatedBenchmarkResult.from_benchmark_results([result, result])



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.plot.helper import arm_name_to_sort_key, extend_range
from ax.utils.common.testutils import TestCase


class HelperTest(TestCase):
    def test_extend_range(self) -> None:
        with self.assertRaises(ValueError):
            extend_range(lower=1, upper=-1)
        self.assertEqual(extend_range(lower=-1, upper=1), (-1.2, 1.2))
        self.assertEqual(extend_range(lower=-1, upper=0, percent=30), (-1.3, 0.3))
        self.assertEqual(extend_range(lower=0, upper=1, percent=50), (-0.5, 1.5))

    def test_arm_name_to_sort_key(self) -> None:
        arm_names = ["0_0", "1_10", "1_2", "10_0", "control"]
        sorted_names = sorted(arm_names, key=arm_name_to_sort_key, reverse=True)
        expected = ["control", "0_0", "1_2", "1_10", "10_0"]
        self.assertEqual(sorted_names, expected)

        arm_names = ["0_0", "0", "1_10", "3_2_x", "3_x", "1_2", "control"]
        sorted_names = sorted(arm_names, key=arm_name_to_sort_key, reverse=True)
        expected = ["control", "3_x", "3_2_x", "0", "0_0", "1_2", "1_10"]
        self.assertEqual(sorted_names, expected)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import enum
from numbers import Real
from typing import List, Tuple

# type aliases
TRGB = Tuple[Real, ...]


class COLORS(enum.Enum):
    STEELBLUE = (128, 177, 211)
    CORAL = (251, 128, 114)
    TEAL = (141, 211, 199)
    PINK = (188, 128, 189)
    LIGHT_PURPLE = (190, 186, 218)
    ORANGE = (253, 180, 98)


# colors to be used for plotting discrete series
# pyre-fixme[5]: Global expression must be annotated.
DISCRETE_COLOR_SCALE = [
    COLORS.STEELBLUE.value,
    COLORS.CORAL.value,
    COLORS.PINK.value,
    COLORS.LIGHT_PURPLE.value,
    COLORS.ORANGE.value,
    COLORS.TEAL.value,
]

# 11-class PiYG from ColorBrewer (for contour plots)
GREEN_PINK_SCALE = [
    (142, 1, 82),
    (197, 27, 125),
    (222, 119, 174),
    (241, 182, 218),
    (253, 224, 239),
    (247, 247, 247),
    (230, 245, 208),
    (184, 225, 134),
    (127, 188, 65),
    (77, 146, 33),
    (39, 100, 25),
]
GREEN_SCALE = [
    (247, 252, 253),
    (229, 245, 249),
    (204, 236, 230),
    (153, 216, 201),
    (102, 194, 164),
    (65, 174, 118),
    (35, 139, 69),
    (0, 109, 44),
    (0, 68, 27),
]
BLUE_SCALE = [
    (255, 247, 251),
    (236, 231, 242),
    (208, 209, 230),
    (166, 189, 219),
    (116, 169, 207),
    (54, 144, 192),
    (5, 112, 176),
    (3, 78, 123),
]
# 24 Class Mixed Color Palette
# Source: https://graphicdesign.stackexchange.com/a/3815
MIXED_SCALE = [
    (2, 63, 165),
    (125, 135, 185),
    (190, 193, 212),
    (214, 188, 192),
    (187, 119, 132),
    (142, 6, 59),
    (74, 111, 227),
    (133, 149, 225),
    (181, 187, 227),
    (230, 175, 185),
    (224, 123, 145),
    (211, 63, 106),
    (17, 198, 56),
    (141, 213, 147),
    (198, 222, 199),
    (234, 211, 198),
    (240, 185, 141),
    (239, 151, 8),
    (15, 207, 192),
    (156, 222, 214),
    (213, 234, 231),
    (243, 225, 235),
    (246, 196, 225),
    (247, 156, 212),
]


def rgba(rgb_tuple: TRGB, alpha: float = 1) -> str:
    """Convert RGB tuple to an RGBA string."""
    return "rgba({},{},{},{alpha})".format(*rgb_tuple, alpha=alpha)


def plotly_color_scale(
    list_of_rgb_tuples: List[TRGB],
    reverse: bool = False,
    alpha: float = 1,
) -> List[Tuple[float, str]]:
    """Convert list of RGB tuples to list of tuples, where each tuple is
    break in [0, 1] and stringified RGBA color.
    """
    if reverse:
        list_of_rgb_tuples = list_of_rgb_tuples[::-1]
    return [
        (round(i / (len(list_of_rgb_tuples) - 1), 3), rgba(rgb))
        for i, rgb in enumerate(list_of_rgb_tuples)
    ]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import warnings
from unittest import mock
from unittest.mock import MagicMock, Mock, patch

from ax.core.observation import ObservationFeatures
from ax.exceptions.core import UserInputError
from ax.modelbridge.factory import get_sobol
from ax.modelbridge.model_spec import FactoryFunctionModelSpec, ModelSpec
from ax.modelbridge.modelbridge_utils import extract_search_space_digest
from ax.modelbridge.registry import Models
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import not_none
from ax.utils.testing.core_stubs import get_branin_experiment
from ax.utils.testing.mock import fast_botorch_optimize


class BaseModelSpecTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.experiment = get_branin_experiment()
        sobol = Models.SOBOL(search_space=self.experiment.search_space)
        sobol_run = sobol.gen(n=20)
        self.experiment.new_batch_trial().add_generator_run(
            sobol_run
        ).run().mark_completed()
        self.data = self.experiment.fetch_data()


class ModelSpecTest(BaseModelSpecTest):
    @fast_botorch_optimize
    def test_construct(self) -> None:
        ms = ModelSpec(model_enum=Models.GPEI)
        with self.assertRaises(UserInputError):
            ms.gen(n=1)
        ms.fit(experiment=self.experiment, data=self.data)
        ms.gen(n=1)

    @fast_botorch_optimize
    # We can use `extract_search_space_digest` as a surrogate for executing
    # the full TorchModelBridge._fit.
    @mock.patch(
        "ax.modelbridge.torch.extract_search_space_digest",
        wraps=extract_search_space_digest,
    )
    def test_fit(self, wrapped_extract_ssd: Mock) -> None:
        ms = ModelSpec(model_enum=Models.GPEI)
        # This should fit the model as usual.
        ms.fit(experiment=self.experiment, data=self.data)
        wrapped_extract_ssd.assert_called_once()
        self.assertIsNotNone(ms._last_fit_arg_ids)
        self.assertEqual(ms._last_fit_arg_ids["experiment"], id(self.experiment))
        # This should skip the model fit.
        with mock.patch("ax.modelbridge.torch.logger") as mock_logger:
            ms.fit(experiment=self.experiment, data=self.data)
        mock_logger.debug.assert_called_with(
            "The observations are identical to the last set of observations "
            "used to fit the model. Skipping model fitting."
        )
        wrapped_extract_ssd.assert_called_once()

    def test_model_key(self) -> None:
        ms = ModelSpec(model_enum=Models.GPEI)
        self.assertEqual(ms.model_key, "GPEI")

    @patch(f"{ModelSpec.__module__}.compute_diagnostics")
    @patch(f"{ModelSpec.__module__}.cross_validate", return_value=["fake-cv-result"])
    def test_cross_validate_with_GP_model(
        self, mock_cv: Mock, mock_diagnostics: Mock
    ) -> None:
        mock_enum = Mock()
        fake_mb = MagicMock()
        fake_mb._process_and_transform_data = MagicMock(return_value=(None, None))
        mock_enum.return_value = fake_mb
        ms = ModelSpec(model_enum=mock_enum, model_cv_kwargs={"test_key": "test-value"})
        ms.fit(
            experiment=self.experiment,
            data=self.experiment.trials[0].fetch_data(),
        )
        cv_results, cv_diagnostics = ms.cross_validate()
        mock_cv.assert_called_with(model=fake_mb, test_key="test-value")
        mock_diagnostics.assert_called_with(["fake-cv-result"])

        self.assertIsNotNone(cv_results)
        self.assertIsNotNone(cv_diagnostics)

        with self.subTest("it caches CV results"):
            mock_cv.reset_mock()
            mock_diagnostics.reset_mock()

            cv_results, cv_diagnostics = ms.cross_validate()

            self.assertIsNotNone(cv_results)
            self.assertIsNotNone(cv_diagnostics)
            mock_cv.assert_not_called()
            mock_diagnostics.assert_not_called()
            self.assertEqual(ms._last_cv_kwargs, {"test_key": "test-value"})

        with self.subTest("fit clears the CV cache"):
            mock_cv.reset_mock()
            mock_diagnostics.reset_mock()

            ms.fit(
                experiment=self.experiment,
                data=self.experiment.trials[0].fetch_data(),
            )
            cv_results, cv_diagnostics = ms.cross_validate()

            self.assertIsNotNone(cv_results)
            self.assertIsNotNone(cv_diagnostics)
            mock_cv.assert_called_with(model=fake_mb, test_key="test-value")
            mock_diagnostics.assert_called_with(["fake-cv-result"])

        with self.subTest("pass in optional kwargs"):
            mock_cv.reset_mock()
            mock_diagnostics.reset_mock()
            # Cache is not empty, but CV will be called since there are new kwargs.
            assert ms._cv_results is not None

            cv_results, cv_diagnostics = ms.cross_validate(model_cv_kwargs={"test": 1})

            self.assertIsNotNone(cv_results)
            self.assertIsNotNone(cv_diagnostics)
            mock_cv.assert_called_with(model=fake_mb, test_key="test-value", test=1)
            self.assertEqual(ms._last_cv_kwargs, {"test": 1, "test_key": "test-value"})

    @patch(f"{ModelSpec.__module__}.compute_diagnostics")
    @patch(f"{ModelSpec.__module__}.cross_validate", side_effect=NotImplementedError)
    def test_cross_validate_with_non_GP_model(
        self, mock_cv: Mock, mock_diagnostics: Mock
    ) -> None:
        mock_enum = Mock()
        mock_enum.return_value = "fake-modelbridge"
        ms = ModelSpec(model_enum=mock_enum, model_cv_kwargs={"test_key": "test-value"})
        ms.fit(
            experiment=self.experiment,
            data=self.experiment.trials[0].fetch_data(),
        )
        with warnings.catch_warnings(record=True) as w:
            cv_results, cv_diagnostics = ms.cross_validate()

        self.assertEqual(len(w), 1)
        self.assertIn("cannot be cross validated", str(w[0].message))
        self.assertIsNone(cv_results)
        self.assertIsNone(cv_diagnostics)

        mock_cv.assert_called_with(model="fake-modelbridge", test_key="test-value")
        mock_diagnostics.assert_not_called()

    def test_fixed_features(self) -> None:
        ms = ModelSpec(model_enum=Models.GPEI)
        self.assertIsNone(ms.fixed_features)
        new_features = ObservationFeatures(parameters={"a": 1.0})
        ms.fixed_features = new_features
        self.assertEqual(ms.fixed_features, new_features)
        self.assertEqual(ms.model_gen_kwargs["fixed_features"], new_features)

    def test_gen_attaches_empty_model_fit_metadata_if_fit_not_applicable(self) -> None:
        ms = ModelSpec(model_enum=Models.SOBOL)
        ms.fit(experiment=self.experiment, data=self.data)
        gr = ms.gen(n=1)
        gen_metadata = not_none(gr.gen_metadata)
        self.assertEqual(gen_metadata["model_fit_quality"], None)
        self.assertEqual(gen_metadata["model_std_quality"], None)
        self.assertEqual(gen_metadata["model_fit_generalization"], None)
        self.assertEqual(gen_metadata["model_std_generalization"], None)

    def test_gen_attaches_model_fit_metadata_if_applicable(self) -> None:
        ms = ModelSpec(model_enum=Models.GPEI)
        ms.fit(experiment=self.experiment, data=self.data)
        gr = ms.gen(n=1)
        gen_metadata = not_none(gr.gen_metadata)
        self.assertIsInstance(gen_metadata["model_fit_quality"], float)
        self.assertIsInstance(gen_metadata["model_std_quality"], float)
        self.assertIsInstance(gen_metadata["model_fit_generalization"], float)
        self.assertIsInstance(gen_metadata["model_std_generalization"], float)


class FactoryFunctionModelSpecTest(BaseModelSpecTest):
    def test_construct(self) -> None:
        ms = FactoryFunctionModelSpec(factory_function=get_sobol)
        with self.assertRaises(UserInputError):
            ms.gen(n=1)
        ms.fit(experiment=self.experiment, data=self.data)
        ms.gen(n=1)

    def test_model_key(self) -> None:
        ms = FactoryFunctionModelSpec(factory_function=get_sobol)
        self.assertEqual(ms.model_key, "get_sobol")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Dict, List, Optional, Tuple

import numpy as np
from ax.core.types import TGenMetadata, TParamValue, TParamValueList
from ax.models.base import Model
from ax.models.types import TConfig


class DiscreteModel(Model):
    """This class specifies the interface for a model based on discrete parameters.

    These methods should be implemented to have access to all of the features
    of Ax.
    """

    def fit(
        self,
        Xs: List[List[TParamValueList]],
        Ys: List[List[float]],
        Yvars: List[List[float]],
        parameter_values: List[TParamValueList],
        outcome_names: List[str],
    ) -> None:
        """Fit model to m outcomes.

        Args:
            Xs: A list of m lists X of parameterizations (each parameterization
                is a list of parameter values of length d), each of length k_i,
                for each outcome.
            Ys: The corresponding list of m lists Y, each of length k_i, for
                each outcome.
            Yvars: The variances of each entry in Ys, same shape.
            parameter_values: A list of possible values for each parameter.
            outcome_names: A list of m outcome names.
        """
        pass

    def predict(self, X: List[TParamValueList]) -> Tuple[np.ndarray, np.ndarray]:
        """Predict

        Args:
            X: List of the j parameterizations at which to make predictions.

        Returns:
            2-element tuple containing

            - (j x m) array of outcome predictions at X.
            - (j x m x m) array of predictive covariances at X.
              cov[j, m1, m2] is Cov[m1@j, m2@j].
        """
        raise NotImplementedError

    def gen(
        self,
        n: int,
        parameter_values: List[TParamValueList],
        objective_weights: Optional[np.ndarray],
        outcome_constraints: Optional[Tuple[np.ndarray, np.ndarray]] = None,
        fixed_features: Optional[Dict[int, TParamValue]] = None,
        pending_observations: Optional[List[List[TParamValueList]]] = None,
        model_gen_options: Optional[TConfig] = None,
    ) -> Tuple[List[TParamValueList], List[float], TGenMetadata]:
        """
        Generate new candidates.

        Args:
            n: Number of candidates to generate.
            parameter_values: A list of possible values for each parameter.
            objective_weights: The objective is to maximize a weighted sum of
                the columns of f(x). These are the weights.
            outcome_constraints: A tuple of (A, b). For k outcome constraints
                and m outputs at f(x), A is (k x m) and b is (k x 1) such that
                A f(x) <= b.
            fixed_features: A map {feature_index: value} for features that
                should be fixed to a particular value during generation.
            pending_observations:  A list of m lists of parameterizations
                (each parameterization is a list of parameter values of length d),
                each of length k_i, for each outcome i.
            model_gen_options: A config dictionary that can contain
                model-specific options.

        Returns:
            2-element tuple containing

            - List of n generated points, where each point is represented
              by a list of parameter values.
            - List of weights for each of the n points.
        """
        raise NotImplementedError

    def cross_validate(
        self,
        Xs_train: List[List[TParamValueList]],
        Ys_train: List[List[float]],
        Yvars_train: List[List[float]],
        X_test: List[TParamValueList],
        use_posterior_predictive: bool = False,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Do cross validation with the given training and test sets.

        Training set is given in the same format as to fit. Test set is given
        in the same format as to predict.

        Args:
            Xs_train: A list of m lists X of parameterizations (each parameterization
                is a list of parameter values of length d), each of length k_i,
                for each outcome.
            Ys_train: The corresponding list of m lists Y, each of length k_i, for
                each outcome.
            Yvars_train: The variances of each entry in Ys, same shape.
            X_test: List of the j parameterizations at which to make predictions.
            use_posterior_predictive: A boolean indicating if the predictions
                should be from the posterior predictive (i.e. including
                observation noise).

        Returns:
            2-element tuple containing

            - (j x m) array of outcome predictions at X.
            - (j x m x m) array of predictive covariances at X.
              `cov[j, m1, m2]` is `Cov[m1@j, m2@j]`.
        """
        raise NotImplementedError

    def best_point(
        self,
        n: int,
        parameter_values: List[TParamValueList],
        objective_weights: Optional[np.ndarray],
        outcome_constraints: Optional[Tuple[np.ndarray, np.ndarray]] = None,
        fixed_features: Optional[Dict[int, TParamValue]] = None,
        pending_observations: Optional[List[List[TParamValueList]]] = None,
        model_gen_options: Optional[TConfig] = None,
    ) -> Optional[TParamValueList]:
        """Obtains the point that has the best value according to the model
        prediction and its model predictions.

        Returns:
            (1 x d) parameter value list representing the point with the best
            value according to the model prediction. None if this function
            is not implemented for the given model.
        """
        return None



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from typing import List, Tuple
from unittest.mock import Mock

import numpy as np
from ax.core import BatchTrial
from ax.core.observation import (
    Observation,
    ObservationData,
    ObservationFeatures,
    observations_from_data,
    recombine_observations,
)
from ax.core.outcome_constraint import OutcomeConstraint
from ax.core.types import ComparisonOp
from ax.metrics.branin import BraninMetric
from ax.modelbridge import ModelBridge
from ax.modelbridge.registry import Models
from ax.modelbridge.transforms.base import Transform
from ax.modelbridge.transforms.relativize import (
    BaseRelativize,
    Relativize,
    RelativizeWithConstantControl,
)
from ax.models.base import Model
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast, not_none
from ax.utils.stats.statstools import relativize_data
from ax.utils.testing.core_stubs import (
    get_branin_data_batch,
    get_branin_experiment,
    get_branin_multi_objective_optimization_config,
    get_branin_optimization_config,
    get_branin_with_multi_task,
    get_search_space,
)
from hypothesis import assume, given, settings, strategies as st


class RelativizeDataTest(TestCase):
    def test_relativize_transform_requires_a_modelbridge(self) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            with self.assertRaisesRegex(
                ValueError, "Relativize transform requires a modelbridge"
            ):
                relativize_cls(
                    search_space=None,
                    observations=[],
                )

    def test_relativize_transform_requires_a_modelbridge_to_have_status_quo_data(
        self,
    ) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            # modelbridge has no status quo
            sobol = Models.SOBOL(search_space=get_search_space())
            self.assertIsNone(sobol.status_quo)
            with self.assertRaisesRegex(
                ValueError, "Cannot relativize data without status quo data"
            ):
                relativize_cls(
                    search_space=None,
                    observations=[],
                    modelbridge=sobol,
                ).transform_observations(
                    observations=[
                        Observation(
                            data=ObservationData(
                                metric_names=["foo"],
                                means=np.array([2]),
                                covariance=np.array([[0.1]]),
                            ),
                            features=ObservationFeatures(parameters={"x": 1}),
                            arm_name="0_0",
                        )
                    ],
                )

            # modelbridge has status quo
            exp = get_branin_experiment(
                with_batch=True,
                with_status_quo=True,
            )
            # making status_quo out of design
            not_none(exp._status_quo)._parameters["x1"] = 10000.0
            for t in exp.trials.values():
                t.mark_running(no_runner_required=True)
                exp.attach_data(
                    get_branin_data_batch(batch=checked_cast(BatchTrial, t))
                )
                t.mark_completed()
            data = exp.fetch_data()
            modelbridge = ModelBridge(
                search_space=exp.search_space,
                model=Model(),
                transforms=[relativize_cls],
                experiment=exp,
                data=data,
            )
            mean_in_data = data.df.query(
                f"arm_name == '{not_none(exp.status_quo).name}'"
            )["mean"].item()
            # modelbridge.status_quo_data_by_trial is accurate
            self.assertEqual(
                mean_in_data, not_none(modelbridge.status_quo_data_by_trial)[0].means[0]
            )

            # create a new experiment
            new_exp = get_branin_experiment(
                with_batch=True,
                with_status_quo=True,
            )
            for t in new_exp.trials.values():
                t.mark_running(no_runner_required=True)
                new_exp.attach_data(
                    get_branin_data_batch(batch=checked_cast(BatchTrial, t))
                )
                t.mark_completed()
            new_data = new_exp.fetch_data()
            new_observations = observations_from_data(experiment=new_exp, data=new_data)
            # calls modelbridge._set_training_data inside
            modelbridge._set_training_data(
                observations=new_observations, search_space=new_exp.search_space
            )
            # The new data is different from the original data
            self.assertNotEqual(data, new_data)
            self.assertFalse(data.df.equals(new_data.df))
            mean_in_data = new_data.df.query(
                f"arm_name == '{not_none(new_exp.status_quo).name}'"
            )["mean"].item()
            # modelbridge.status_quo_data_by_trial remains accurate
            self.assertEqual(
                mean_in_data, not_none(modelbridge.status_quo_data_by_trial)[0].means[0]
            )

            # Can still find status_quo_data_by_trial when status_quo name is None
            mb_sq = not_none(modelbridge._status_quo)
            mb_sq.arm_name = None
            self.assertIsNotNone(modelbridge.status_quo_data_by_trial)

            # test transform edge cases
            observations = observations_from_data(
                experiment=exp,
                data=data,
            )
            tf = relativize_cls(
                search_space=None,
                observations=observations,
                modelbridge=modelbridge,
            )
            # making observation coming from trial_index not in modelbridge
            observations[0].features.trial_index = 999
            self.assertRaises(ValueError, tf.transform_observations, observations)

    def test_relativize_transform_observations(self) -> None:
        def _check_transform_observations(
            tf: Transform,
            observations: List[Observation],
            expected_mean_and_covar: List[Tuple[np.ndarray, np.ndarray]],
        ) -> None:
            results = tf.transform_observations(observations)
            for i, tsfm_obs in enumerate(results):
                expected_mean, expected_covar = expected_mean_and_covar[i]
                self.assertEqual(tsfm_obs.data.metric_names, metric_names)
                # status quo means must always be zero
                self.assertTrue(
                    np.allclose(tsfm_obs.data.means, expected_mean),
                    tsfm_obs.data.means,
                )
                # status quo covariances must always be zero
                self.assertTrue(
                    np.allclose(tsfm_obs.data.covariance, expected_covar),
                    tsfm_obs.data.covariance,
                )
            # Check untransform
            untsfm_results = tf.untransform_observations(results)
            for i, untsfm_obs in enumerate(untsfm_results):
                obs = observations[i]
                self.assertTrue(np.allclose(untsfm_obs.data.means, obs.data.means))
                self.assertTrue(
                    np.allclose(untsfm_obs.data.covariance, obs.data.covariance)
                )

        metric_names = ["foobar", "foobaz"]
        arm_names = ["status_quo", "0_0"]
        obs_data = [
            ObservationData(
                metric_names=metric_names,
                means=np.array([2.0, 5.0]),
                covariance=np.array([[0.1, 0.0], [0.0, 0.2]]),
            ),
            ObservationData(
                metric_names=metric_names,
                means=np.array([1.0, 10.0]),
                covariance=np.array([[0.3, 0.0], [0.0, 0.4]]),
            ),
        ]
        obs_features = [
            ObservationFeatures(parameters={"x": 1}, trial_index=0),
            ObservationFeatures(parameters={"x": 2}, trial_index=0),
        ]

        observations = recombine_observations(obs_features, obs_data, arm_names)
        modelbridge = Mock(
            status_quo=Mock(
                data=obs_data[0], features=obs_features[0], arm_name=arm_names[0]
            ),
            status_quo_data_by_trial={0: obs_data[0]},
        )

        for relativize_cls, expected_mean_and_covar in [
            (
                Relativize,
                [
                    (np.array([0.0, 0.0]), np.array([[0.0, 0.0], [0.0, 0.0]])),
                    (np.array([-51.25, 98.4]), np.array([[812.5, 0.0], [0.0, 480.0]])),
                ],
            ),
            (
                RelativizeWithConstantControl,
                [
                    (np.array([0.0, 0.0]), np.array([[0.0, 0.0], [0.0, 0.0]])),
                    (np.array([-50.0, 100.0]), np.array([[750.0, 0.0], [0.0, 160.0]])),
                ],
            ),
        ]:
            tf = relativize_cls(
                search_space=None,
                observations=observations,
                modelbridge=modelbridge,
            )
            # check transform and untransform on observations
            _check_transform_observations(
                tf=tf,
                observations=observations,
                expected_mean_and_covar=expected_mean_and_covar,
            )
            # transform should still work when trial_index is None
            modelbridge = Mock(
                status_quo=Mock(
                    data=obs_data[0], features=obs_features[0], arm_name=arm_names[0]
                ),
                status_quo_data_by_trial={0: obs_data[1], 1: obs_data[0]},
            )
            tf = relativize_cls(
                search_space=None,
                observations=observations,
                modelbridge=modelbridge,
            )
            for obs in observations:
                obs.features.trial_index = None
            _check_transform_observations(
                tf=tf,
                observations=observations,
                expected_mean_and_covar=expected_mean_and_covar,
            )

        # Check instantiation and subclassing of BaseRelativize
        class BadRelativize(BaseRelativize):
            pass

        for abstract_cls in [BaseRelativize, BadRelativize]:
            with self.assertRaisesRegex(TypeError, "Can't instantiate abstract class"):
                abstract_cls(
                    search_space=None,
                    observations=observations,
                    modelbridge=modelbridge,
                )

    # pyre-fixme[56]: Pyre was not able to infer the type of argument
    #  `hypothesis.strategies.floats($parameter$min_value = - 10.000000,
    #  $parameter$max_value = 10.000000)` to decorator factory `hypothesis.given`.
    @given(
        st.floats(min_value=-10.0, max_value=10.0),
        st.floats(min_value=0, max_value=10.0),
        st.floats(min_value=-10.0, max_value=10.0),
        st.floats(min_value=0, max_value=10.0),
    )
    @settings(max_examples=1000, deadline=None)
    def test_transform_status_quos_always_zero(
        self,
        sq_mean: float,
        sq_sem: float,
        mean: float,
        sem: float,
    ) -> None:
        assume(abs(sq_mean) >= 1e-10)
        assume(abs(sq_mean) != sq_sem)

        arm_names = ["status_quo", "0_0"]
        obs_data = [
            ObservationData(
                metric_names=["foo"],
                means=np.array([sq_mean]),
                covariance=np.array([[sq_sem]]),
            ),
            ObservationData(
                metric_names=["foo"],
                means=np.array([mean]),
                covariance=np.array([[sem]]),
            ),
        ]
        obs_features = [
            ObservationFeatures(parameters={"x": 1}, trial_index=0),
            ObservationFeatures(parameters={"x": 2}, trial_index=0),
        ]
        modelbridge = Mock(
            status_quo=Mock(
                data=obs_data[0], features=obs_features[0], arm_name=arm_names[0]
            ),
            status_quo_data_by_trial={0: obs_data[0]},
        )
        observations = recombine_observations(obs_features, obs_data, arm_names)
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            transform = relativize_cls(
                search_space=None,
                observations=observations,
                modelbridge=modelbridge,
            )
            relative_obs = transform.transform_observations(observations)
            self.assertEqual(relative_obs[0].data.metric_names, ["foo"])
            self.assertAlmostEqual(relative_obs[0].data.means[0], 0, places=4)
            self.assertAlmostEqual(relative_obs[0].data.covariance[0][0], 0, places=4)

    def test_multitask_data(self) -> None:
        experiment = get_branin_with_multi_task()
        data = experiment.fetch_data()

        observations = observations_from_data(
            experiment=experiment,
            data=data,
        )
        relative_observations = observations_from_data(
            experiment=experiment,
            data=relativize_data(
                data=data,
                status_quo_name="status_quo",
                as_percent=True,
                include_sq=True,
            ),
        )

        sq_obs_data = []
        for i in data.df["trial_index"].unique():
            status_quo_data = data.df.loc[
                (data.df["arm_name"] == "status_quo") & (data.df["trial_index"] == i)
            ]
            sq_obs_data.append(
                ObservationData(
                    metric_names=status_quo_data["metric_name"].to_numpy(),
                    means=status_quo_data["mean"].to_numpy(),
                    covariance=status_quo_data["sem"].to_numpy()[np.newaxis, :] ** 2,
                )
            )

        modelbridge = Mock(
            status_quo=Observation(
                data=sq_obs_data[0],
                features=ObservationFeatures(
                    parameters=not_none(experiment.status_quo).parameters
                ),
                arm_name="status_quo",
            ),
            status_quo_data_by_trial={
                i: sq_obs_data[i] for i in range(len(sq_obs_data))
            },
        )

        # not checking RelativizeWithConstantControl here
        # because relativize_data uses delta method
        transform = Relativize(
            search_space=None,
            observations=observations,
            modelbridge=modelbridge,
        )

        relative_obs_t = transform.transform_observations(observations)
        self.maxDiff = None
        # this assertion just checks that order is the same, which
        # is only important for the purposes of this test
        self.assertEqual(
            [datum.data.metric_names for datum in relative_obs_t],
            [datum.data.metric_names for datum in relative_observations],
        )
        means = [
            np.array([datum.data.means for datum in relative_obs_t]),
            np.array([datum.data.means for datum in relative_observations]),
        ]
        # `self.assertAlmostEqual(relative_obs_data, expected_obs_data)`
        # fails 1% of the time, so we check with numpy.
        self.assertTrue(
            all(np.isclose(means[0], means[1])),
            means,
        )
        covariances = [
            np.array([datum.data.covariance for datum in relative_observations]),
            np.array([datum.data.covariance for datum in relative_obs_t]),
        ]
        self.assertTrue(
            all(np.isclose(covariances[0], covariances[1])),
            covariances,
        )


class RelativizeDataOptConfigTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        search_space = get_search_space()
        gr = Models.SOBOL(search_space=search_space).gen(n=1)
        self.model = Mock(
            search_space=search_space,
            status_quo=Mock(
                features=ObservationFeatures(parameters=gr.arms[0].parameters)
            ),
        )

    def test_transform_optimization_config_without_constraints(self) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            relativize = relativize_cls(
                search_space=None,
                observations=[],
                modelbridge=self.model,
            )
            optimization_config = get_branin_optimization_config()
            new_config = relativize.transform_optimization_config(
                optimization_config=optimization_config,
                modelbridge=None,
                fixed_features=Mock(),
            )
            self.assertEqual(new_config.objective, optimization_config.objective)

    def test_transform_optimization_config_with_relative_constraints(self) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            relativize = relativize_cls(
                search_space=None,
                observations=[],
                modelbridge=self.model,
            )
            optimization_config = get_branin_optimization_config()
            optimization_config.outcome_constraints = [
                OutcomeConstraint(
                    metric=BraninMetric("b2", ["x2", "x1"]),
                    op=ComparisonOp.GEQ,
                    bound=-200.0,
                    relative=True,
                )
            ]
            new_config = relativize.transform_optimization_config(
                optimization_config=deepcopy(optimization_config),
                modelbridge=None,
                fixed_features=Mock(),
            )
            self.assertEqual(new_config.objective, optimization_config.objective)
            self.assertEqual(
                new_config.outcome_constraints[0].bound,
                optimization_config.outcome_constraints[0].bound,
            )
            self.assertFalse(new_config.outcome_constraints[0].relative)
            # Untransform the constraints
            cons = relativize.untransform_outcome_constraints(
                outcome_constraints=new_config.outcome_constraints,
                fixed_features=Mock(),
            )
            self.assertEqual(cons, optimization_config.outcome_constraints)

    def test_transform_optimization_config_with_non_relative_constraints(self) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            relativize = relativize_cls(
                search_space=None,
                observations=[],
                modelbridge=self.model,
            )
            optimization_config = get_branin_optimization_config()
            optimization_config.outcome_constraints = [
                OutcomeConstraint(
                    metric=BraninMetric("b2", ["x2", "x1"]),
                    op=ComparisonOp.GEQ,
                    bound=-200.0,
                    relative=False,
                )
            ]
            with self.assertRaisesRegex(ValueError, "All constraints must be relative"):
                relativize.transform_optimization_config(
                    optimization_config=optimization_config,
                    modelbridge=None,
                    fixed_features=Mock(),
                )

    def test_transform_optimization_config_with_relative_thresholds(self) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            relativize = relativize_cls(
                search_space=None,
                observations=[],
                modelbridge=self.model,
            )
            optimization_config = get_branin_multi_objective_optimization_config(
                has_objective_thresholds=True,
            )
            for threshold in optimization_config.objective_thresholds:
                threshold.relative = True

            new_config = relativize.transform_optimization_config(
                optimization_config=optimization_config,
                modelbridge=None,
                fixed_features=Mock(),
            )
            self.assertEqual(new_config.objective, optimization_config.objective)
            self.assertEqual(
                # pyre-fixme[16]: `OptimizationConfig` has no attribute
                #  `objective_thresholds`.
                new_config.objective_thresholds[0].bound,
                optimization_config.objective_thresholds[0].bound,
            )
            self.assertFalse(new_config.objective_thresholds[0].relative)
            self.assertEqual(
                new_config.objective_thresholds[1].bound,
                optimization_config.objective_thresholds[1].bound,
            )
            self.assertFalse(new_config.objective_thresholds[1].relative)

    def test_transform_optimization_config_with_non_relative_thresholds(self) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            relativize = relativize_cls(
                search_space=None,
                observations=[],
                modelbridge=self.model,
            )
            optimization_config = get_branin_multi_objective_optimization_config(
                has_objective_thresholds=True,
            )
            optimization_config.objective_thresholds[1].relative = False

            with self.assertRaisesRegex(
                ValueError, "All objective thresholds must be relative"
            ):
                relativize.transform_optimization_config(
                    optimization_config=optimization_config,
                    modelbridge=None,
                    fixed_features=Mock(),
                )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, List

from ax.core.base_trial import BaseTrial
from ax.core.map_metric import MapMetricFetchResult
from ax.metrics.noisy_function_map import NoisyFunctionMapMetric


class BackendSimulatorTimestampMapMetric(NoisyFunctionMapMetric):
    """A metric that interfaces with an underlying ``BackendSimulator`` and
    returns timestamp map data."""

    def fetch_trial_data(
        self, trial: BaseTrial, noisy: bool = True, **kwargs: Any
    ) -> MapMetricFetchResult:
        """Fetch data for one trial."""
        backend_simulator = trial.experiment.runner.simulator  # pyre-ignore[16]
        sim_trial = backend_simulator.get_sim_trial_by_index(trial.index)
        end_time = (
            backend_simulator.time
            if sim_trial.sim_completed_time is None
            else sim_trial.sim_completed_time
        )
        timestamps = self.convert_to_timestamps(
            start_time=sim_trial.sim_start_time, end_time=end_time
        )
        timestamp_kwargs = {"map_keys": ["timestamp"], "timestamp": timestamps}
        return NoisyFunctionMapMetric.fetch_trial_data(
            self, trial=trial, noisy=noisy, **kwargs, **timestamp_kwargs
        )

    def convert_to_timestamps(self, start_time: float, end_time: float) -> List[float]:
        """Given a starting and current time, get the list of intermediate
        timestamps at which we have observations."""
        raise NotImplementedError



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections import OrderedDict
from typing import List
from unittest import mock

import numpy as np
from ax.core.observation import ObservationFeatures
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.parameter_constraint import (
    OrderConstraint,
    ParameterConstraint,
    SumConstraint,
)
from ax.core.search_space import SearchSpace
from ax.exceptions.core import SearchSpaceExhausted
from ax.modelbridge.random import RandomModelBridge
from ax.modelbridge.registry import Cont_X_trans
from ax.models.random.base import RandomModel
from ax.models.random.sobol import SobolGenerator
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_small_discrete_search_space


class RandomModelBridgeTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        x = RangeParameter("x", ParameterType.FLOAT, lower=0, upper=1)
        y = RangeParameter("y", ParameterType.FLOAT, lower=1, upper=2)
        z = RangeParameter("z", ParameterType.FLOAT, lower=0, upper=5)
        self.parameters = [x, y, z]
        parameter_constraints: List[ParameterConstraint] = [
            OrderConstraint(x, y),
            SumConstraint([x, z], False, 3.5),
        ]
        self.search_space = SearchSpace(self.parameters, parameter_constraints)
        self.model_gen_options = {"option": "yes"}

    @mock.patch("ax.modelbridge.random.RandomModelBridge.__init__", return_value=None)
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def test_Fit(self, mock_init):
        # pyre-fixme[20]: Argument `model` expected.
        modelbridge = RandomModelBridge()
        model = mock.create_autospec(RandomModel, instance=True)
        modelbridge._fit(model, self.search_space, None)
        self.assertEqual(modelbridge.parameters, ["x", "y", "z"])
        self.assertTrue(isinstance(modelbridge.model, RandomModel))

    @mock.patch("ax.modelbridge.random.RandomModelBridge.__init__", return_value=None)
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def test_Predict(self, mock_init):
        # pyre-fixme[20]: Argument `model` expected.
        modelbridge = RandomModelBridge()
        modelbridge.transforms = OrderedDict()
        modelbridge.parameters = ["x", "y", "z"]
        with self.assertRaises(NotImplementedError):
            modelbridge._predict([])

    @mock.patch("ax.modelbridge.random.RandomModelBridge.__init__", return_value=None)
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def test_CrossValidate(self, mock_init):
        # pyre-fixme[20]: Argument `model` expected.
        modelbridge = RandomModelBridge()
        modelbridge.transforms = OrderedDict()
        modelbridge.parameters = ["x", "y", "z"]
        with self.assertRaises(NotImplementedError):
            modelbridge._cross_validate(self.search_space, [], [])

    @mock.patch(
        "ax.models.random.base.RandomModel.gen",
        autospec=True,
        return_value=(
            np.array([[1.0, 2.0, 3.0], [3.0, 4.0, 3.0]]),
            np.array([1.0, 2.0]),
        ),
    )
    @mock.patch("ax.modelbridge.random.RandomModelBridge.__init__", return_value=None)
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def test_Gen(self, mock_init, mock_gen):
        # Test with constraints
        # pyre-fixme[20]: Argument `model` expected.
        modelbridge = RandomModelBridge()
        modelbridge.parameters = ["x", "y", "z"]
        modelbridge.transforms = OrderedDict()
        modelbridge.model = RandomModel()
        gen_results = modelbridge._gen(
            n=3,
            search_space=self.search_space,
            pending_observations={},
            fixed_features=ObservationFeatures({"z": 3.0}),
            optimization_config=None,
            # pyre-fixme[6]: For 6th param expected `Optional[Dict[str, Union[None,
            #  Dict[str, typing.Any], OptimizationConfig, AcquisitionFunction, float,
            #  int, str]]]` but got `Dict[str, str]`.
            model_gen_options=self.model_gen_options,
        )
        gen_args = mock_gen.mock_calls[0][2]
        self.assertEqual(gen_args["n"], 3)
        self.assertEqual(gen_args["bounds"], [(0.0, 1.0), (1.0, 2.0), (0.0, 5.0)])
        self.assertTrue(
            np.array_equal(
                gen_args["linear_constraints"][0],
                np.array([[1.0, -1, 0.0], [-1.0, 0.0, -1.0]]),
            )
        )
        self.assertTrue(
            np.array_equal(gen_args["linear_constraints"][1], np.array([[0.0], [-3.5]]))
        )
        self.assertEqual(gen_args["fixed_features"], {2: 3.0})
        self.assertEqual(gen_args["model_gen_options"], {"option": "yes"})
        obsf = gen_results.observation_features
        self.assertEqual(obsf[0].parameters, {"x": 1.0, "y": 2.0, "z": 3.0})
        self.assertEqual(obsf[1].parameters, {"x": 3.0, "y": 4.0, "z": 3.0})
        self.assertTrue(np.array_equal(gen_results.weights, np.array([1.0, 2.0])))

        # Test with no constraints, no fixed feature, no pending observations
        search_space = SearchSpace(self.parameters[:2])
        modelbridge.parameters = ["x", "y"]
        modelbridge._gen(
            n=3,
            search_space=search_space,
            pending_observations={},
            fixed_features=ObservationFeatures({}),
            optimization_config=None,
            # pyre-fixme[6]: For 6th param expected `Optional[Dict[str, Union[None,
            #  Dict[str, typing.Any], OptimizationConfig, AcquisitionFunction, float,
            #  int, str]]]` but got `Dict[str, str]`.
            model_gen_options=self.model_gen_options,
        )
        gen_args = mock_gen.mock_calls[1][2]
        self.assertEqual(gen_args["bounds"], [(0.0, 1.0), (1.0, 2.0)])
        self.assertIsNone(gen_args["linear_constraints"])
        self.assertIsNone(gen_args["fixed_features"])

    def test_deduplicate(self) -> None:
        sobol = RandomModelBridge(
            search_space=get_small_discrete_search_space(),
            model=SobolGenerator(deduplicate=True),
            transforms=Cont_X_trans,
        )
        for _ in range(4):  # Search space is {[0, 1], {"red", "panda"}}
            self.assertEqual(len(sobol.gen(1).arms), 1)
        with self.assertRaises(SearchSpaceExhausted):
            sobol.gen(1)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import random
import string
from typing import Tuple
from unittest.mock import patch

from ax.core.base_trial import TrialStatus
from ax.core.experiment import Experiment
from ax.modelbridge.generation_strategy import GenerationStrategy
from ax.service.utils.with_db_settings_base import WithDBSettingsBase
from ax.storage.sqa_store.db import init_test_engine_and_session_factory
from ax.storage.sqa_store.load import (
    _load_experiment,
    _load_generation_strategy_by_experiment_name,
)
from ax.storage.sqa_store.reduced_state import GR_LARGE_MODEL_ATTRS
from ax.storage.sqa_store.save import (
    _save_experiment,
    _save_generation_strategy,
    _save_or_update_trials,
)
from ax.storage.sqa_store.structs import DBSettings
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_experiment,
    get_generator_run,
    SpecialGenerationStrategy,
)
from ax.utils.testing.modeling_stubs import get_generation_strategy


class TestWithDBSettingsBase(TestCase):
    """Tests saving/loading functionality of WithDBSettingsBase class."""

    def setUp(self) -> None:
        self.generation_strategy = get_generation_strategy(with_experiment=True)
        self.experiment = self.generation_strategy.experiment

        init_test_engine_and_session_factory(force_init=True)
        self.with_db_settings = WithDBSettingsBase(
            db_settings=DBSettings(url="sqlite://")
        )
        _save_experiment(
            self.experiment,
            encoder=self.with_db_settings.db_settings.encoder,
            decoder=self.with_db_settings.db_settings.decoder,
        )
        _save_generation_strategy(
            generation_strategy=self.generation_strategy,
            encoder=self.with_db_settings.db_settings.encoder,
            decoder=self.with_db_settings.db_settings.decoder,
        )

    def get_random_experiment(self) -> Experiment:
        """Get an Experiment instance with random name."""

        experiment = get_experiment()
        experiment_name = "".join(random.choice(string.ascii_letters) for i in range(8))
        experiment.name = experiment_name
        return experiment

    def get_random_generation_strategy(self) -> GenerationStrategy:
        """Get an GenerationStrategy instance with random name."""

        generation_strategy = get_generation_strategy(with_callable_model_kwarg=False)
        gs_name = "".join(random.choice(string.ascii_letters) for i in range(8))
        generation_strategy._name = gs_name
        return generation_strategy

    def init_experiment_and_generation_strategy(
        self, save_experiment: bool = True, save_generation_strategy: bool = True
    ) -> Tuple[Experiment, GenerationStrategy]:
        """Generate a random Experiment and associated generation_strategy"""

        generation_strategy = self.get_random_generation_strategy()
        experiment = self.get_random_experiment()
        generation_strategy.experiment = experiment

        if save_experiment:
            _save_experiment(
                experiment,
                encoder=self.with_db_settings.db_settings.encoder,
                decoder=self.with_db_settings.db_settings.decoder,
            )
        if save_generation_strategy:
            _save_generation_strategy(
                generation_strategy=generation_strategy,
                encoder=self.with_db_settings.db_settings.encoder,
                decoder=self.with_db_settings.db_settings.decoder,
            )
        return experiment, generation_strategy

    def test_get_experiment_and_generation_strategy_db_id(self) -> None:

        (
            exp_id,
            gen_id,
        ) = self.with_db_settings._get_experiment_and_generation_strategy_db_id(
            self.experiment.name
        )
        self.assertIsNotNone(exp_id)
        self.assertIsNotNone(gen_id)

    def test_save_experiment(self) -> None:
        experiment = self.get_random_experiment()
        saved = self.with_db_settings._save_experiment_to_db_if_possible(experiment)
        self.assertTrue(saved)
        loaded_experiment = _load_experiment(
            experiment.name, decoder=self.with_db_settings.db_settings.decoder
        )
        self.assertIsNotNone(loaded_experiment)
        self.assertEqual(experiment, loaded_experiment)

    def test_save_generation_strategy(self) -> None:
        experiment, generation_strategy = self.init_experiment_and_generation_strategy(
            save_generation_strategy=False
        )
        saved = self.with_db_settings._save_generation_strategy_to_db_if_possible(
            generation_strategy
        )
        self.assertTrue(saved)
        loaded_gs = _load_generation_strategy_by_experiment_name(
            experiment_name=experiment.name,
            decoder=self.with_db_settings.db_settings.decoder,
        )
        self.assertIsNotNone(loaded_gs)
        self.assertEqual(loaded_gs.name, generation_strategy.name)

    def test_save_non_standard_generation_strategy(self) -> None:
        generation_strategy = SpecialGenerationStrategy()
        saved = self.with_db_settings._save_generation_strategy_to_db_if_possible(
            generation_strategy
        )
        self.assertFalse(saved)

    def test_save_load_experiment_and_generation_strategy(self) -> None:
        experiment, generation_strategy = self.init_experiment_and_generation_strategy(
            save_generation_strategy=False
        )
        db_exp, db_gs = self.with_db_settings._load_experiment_and_generation_strategy(
            experiment.name
        )
        self.assertIsNotNone(db_exp)
        self.assertIsNone(db_gs)

        experiment, generation_strategy = self.init_experiment_and_generation_strategy(
            save_experiment=False, save_generation_strategy=False
        )
        (
            exp_saved,
            gs_saved,
        ) = self.with_db_settings._maybe_save_experiment_and_generation_strategy(
            experiment, generation_strategy
        )
        self.assertTrue(exp_saved)
        self.assertTrue(gs_saved)

        db_exp, db_gs = self.with_db_settings._load_experiment_and_generation_strategy(
            experiment.name
        )
        self.assertIsNotNone(db_exp)
        self.assertEqual(db_exp, experiment)
        self.assertIsNotNone(db_gs)
        self.assertEqual(db_gs.name, generation_strategy.name)

    def test_update_generation_strategy(self) -> None:
        _, generation_strategy = self.init_experiment_and_generation_strategy()

        generator_run = get_generator_run()
        self.assertIsNone(generator_run.db_id)
        updated = self.with_db_settings._update_generation_strategy_in_db_if_possible(
            generation_strategy, [generator_run]
        )
        self.assertTrue(updated)
        self.assertIsNotNone(generator_run.db_id)
        self.assertIsNotNone(generator_run.arms[0].db_id)

    def test_update_non_standard_generation_strategy(self) -> None:
        generation_strategy = SpecialGenerationStrategy()
        generator_run = get_generator_run()
        saved = self.with_db_settings._update_generation_strategy_in_db_if_possible(
            generation_strategy, [generator_run]
        )
        self.assertFalse(saved)
        self.assertIsNone(generator_run.db_id)
        self.assertIsNone(generator_run.arms[0].db_id)

    @patch(f"{WithDBSettingsBase.__module__}.STORAGE_MINI_BATCH_SIZE", 2)
    def test_update_generation_strategy_mini_batches(self) -> None:
        _, generation_strategy = self.init_experiment_and_generation_strategy()

        # Check with 1 GR.
        generator_run = get_generator_run()
        self.assertIsNone(generator_run.db_id)
        updated = self.with_db_settings._update_generation_strategy_in_db_if_possible(
            generation_strategy, [generator_run]
        )
        self.assertTrue(updated)
        self.assertIsNotNone(generator_run.db_id)

        # Check with multiple GRs, where their number % mini batch size is not 0.
        grs = [generator_run.clone() for _ in range(5)]
        for gr in grs:
            self.assertIsNone(gr._db_id)
        updated = self.with_db_settings._update_generation_strategy_in_db_if_possible(
            generation_strategy, grs
        )
        self.assertTrue(updated)
        for gr in grs:
            self.assertIsNotNone(gr.db_id)

    def test_save_new_trial(self) -> None:
        experiment, _ = self.init_experiment_and_generation_strategy(
            save_generation_strategy=False
        )

        exp = _load_experiment(
            experiment.name, decoder=self.with_db_settings.db_settings.decoder
        )
        trial = exp.new_trial()
        saved = self.with_db_settings._save_or_update_trial_in_db_if_possible(
            exp, trial
        )
        self.assertTrue(saved)
        exp = _load_experiment(
            experiment.name, decoder=self.with_db_settings.db_settings.decoder
        )
        self.assertEqual(len(exp.trials), 1)
        self.assertEqual(exp.trials[0].status, TrialStatus.CANDIDATE)

    def test_save_updated_trial(self) -> None:
        experiment, _ = self.init_experiment_and_generation_strategy(
            save_generation_strategy=False
        )

        exp = _load_experiment(
            experiment.name, decoder=self.with_db_settings.db_settings.decoder
        )
        trial = exp.new_trial()
        _save_or_update_trials(
            experiment=experiment,
            trials=[trial],
            encoder=self.with_db_settings.db_settings.encoder,
            decoder=self.with_db_settings.db_settings.decoder,
        )
        self.assertEqual(trial.status, TrialStatus.CANDIDATE)

        trial.mark_running(True)
        saved = self.with_db_settings._save_or_update_trial_in_db_if_possible(
            exp, trial
        )
        self.assertTrue(saved)
        exp = _load_experiment(
            experiment.name, decoder=self.with_db_settings.db_settings.decoder
        )
        self.assertEqual(len(exp.trials), 1)
        self.assertEqual(exp.trials[0].status, TrialStatus.RUNNING)

    @patch(f"{WithDBSettingsBase.__module__}.STORAGE_MINI_BATCH_SIZE", 2)
    def test_updated_trials_mini_batch(self) -> None:
        experiment, _ = self.init_experiment_and_generation_strategy(
            save_generation_strategy=False
        )

        # Check with 1 trial.
        trial = experiment.new_trial()
        self.assertIsNone(trial.db_id)
        self.with_db_settings._save_or_update_trials_in_db_if_possible(
            experiment=experiment,
            trials=[trial],
        )
        loaded_experiment = _load_experiment(
            experiment.name, decoder=self.with_db_settings.db_settings.decoder
        )
        self.assertEqual(
            # pyre-fixme[16]: Optional type has no attribute `status`.
            loaded_experiment.trials.get(trial.index).status,
            TrialStatus.CANDIDATE,
        )
        self.assertIsNotNone(trial.db_id)

        # Check with multiple trials, where their number % mini batch size is not 0.
        trials = [experiment.new_trial() for _ in range(5)]
        for t in trials:
            self.assertIsNone(t.db_id)

        trial.mark_running(no_runner_required=True)
        trials.append(trial)

        self.with_db_settings._save_or_update_trials_in_db_if_possible(
            experiment=experiment,
            # pyre-fixme[6]: For 2nd param expected `List[BaseTrial]` but got
            #  `List[Trial]`.
            trials=trials,
        )
        loaded_experiment = _load_experiment(
            experiment.name, decoder=self.with_db_settings.db_settings.decoder
        )
        # All trials except for the one we marked as running should be candidates.
        for t in trials:
            self.assertIsNotNone(t.db_id)
            if t.index != trial.index:
                self.assertEqual(t.status, TrialStatus.CANDIDATE)
            else:
                self.assertEqual(t.status, TrialStatus.RUNNING)

    def test_update_reduced_state_generator_runs(self) -> None:
        experiment, generation_strategy = self.init_experiment_and_generation_strategy(
            save_generation_strategy=True
        )

        trials = [experiment.new_trial() for _ in range(5)]
        grs = []
        for t in trials:
            gr = generation_strategy.gen(experiment)
            grs.append(gr)
            t.add_generator_run(gr)

        self.with_db_settings._save_or_update_trials_and_generation_strategy_if_possible(  # noqa E501
            experiment=experiment,
            # pyre-fixme[6]: For 2nd param expected `List[BaseTrial]` but got
            #  `List[Trial]`.
            trials=trials,
            generation_strategy=generation_strategy,
            new_generator_runs=grs,
            reduce_state_generator_runs=True,
        )

        loaded_experiment = _load_experiment(
            experiment.name, decoder=self.with_db_settings.db_settings.decoder
        )

        # Only the last trial's generator run should have large model attributes
        for idx, trial in loaded_experiment.trials.items():
            for key in [f"_{attr.key}" for attr in GR_LARGE_MODEL_ATTRS]:
                if idx < len(loaded_experiment.trials) - 1:
                    # pyre-fixme[16]: `BaseTrial` has no attribute `generator_run`.
                    self.assertIsNone(getattr(trial.generator_run, key))
                else:
                    self.assertIsNotNone(getattr(trial.generator_run, key))

        loaded_generation_strategy = _load_generation_strategy_by_experiment_name(
            experiment.name, decoder=self.with_db_settings.db_settings.decoder
        )

        # Only the last generator run should have large model attributes
        for idx, gr in enumerate(loaded_generation_strategy._generator_runs):
            for key in [f"_{attr.key}" for attr in GR_LARGE_MODEL_ATTRS]:
                if idx < len(loaded_generation_strategy._generator_runs) - 1:
                    self.assertIsNone(getattr(gr, key))
                else:
                    self.assertIsNotNone(getattr(gr, key))

    def test_update_experiment_properties_in_db(self) -> None:
        experiment, _ = self.init_experiment_and_generation_strategy(
            save_generation_strategy=False
        )
        experiment._properties["test_property"] = True
        self.with_db_settings._update_experiment_properties_in_db(
            experiment_with_updated_properties=experiment
        )
        loaded_experiment = _load_experiment(
            experiment.name, decoder=self.with_db_settings.db_settings.decoder
        )
        self.assertEqual(loaded_experiment._properties, {"test_property": True})



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import inspect
import logging
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Type, TYPE_CHECKING, Union

import torch
from ax.core.arm import Arm
from ax.core.base_trial import TrialStatus
from ax.core.batch_trial import (
    AbandonedArm,
    BatchTrial,
    GeneratorRunStruct,
    LifecycleStage,
)
from ax.core.generator_run import GeneratorRun
from ax.core.runner import Runner
from ax.core.trial import Trial
from ax.exceptions.storage import JSONDecodeError
from ax.modelbridge.transforms.base import Transform
from ax.storage.botorch_modular_registry import (
    CLASS_TO_REVERSE_REGISTRY,
    REVERSE_INPUT_TRANSFORM_REGISTRY,
    REVERSE_OUTCOME_TRANSFORM_REGISTRY,
)
from ax.storage.transform_registry import REVERSE_TRANSFORM_REGISTRY
from ax.utils.common.kwargs import warn_on_kwargs
from ax.utils.common.logger import get_logger
from ax.utils.common.typeutils import checked_cast
from ax.utils.common.typeutils_torch import torch_type_from_str
from botorch.models.transforms.input import ChainedInputTransform, InputTransform
from botorch.models.transforms.outcome import ChainedOutcomeTransform, OutcomeTransform
from botorch.utils.types import _DefaultType, DEFAULT

logger: logging.Logger = get_logger(__name__)


if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import core  # noqa F401


def batch_trial_from_json(
    experiment: core.experiment.Experiment,
    index: int,
    trial_type: Optional[str],
    status: TrialStatus,
    time_created: datetime,
    time_completed: Optional[datetime],
    time_staged: Optional[datetime],
    time_run_started: Optional[datetime],
    abandoned_reason: Optional[str],
    run_metadata: Optional[Dict[str, Any]],
    generator_run_structs: List[GeneratorRunStruct],
    runner: Optional[Runner],
    abandoned_arms_metadata: Dict[str, AbandonedArm],
    num_arms_created: int,
    status_quo: Optional[Arm],
    status_quo_weight_override: float,
    optimize_for_power: Optional[bool],
    # Allowing default values for backwards compatibility with
    # objects stored before these fields were added.
    failed_reason: Optional[str] = None,
    ttl_seconds: Optional[int] = None,
    generation_step_index: Optional[int] = None,
    properties: Optional[Dict[str, Any]] = None,
    stop_metadata: Optional[Dict[str, Any]] = None,
    lifecycle_stage: Optional[LifecycleStage] = None,
    **kwargs: Any,
) -> BatchTrial:
    """Load Ax BatchTrial from JSON.

    Other classes don't need explicit deserializers, because we can just use
    their constructors (see decoder.py). However, the constructor for Batch
    does not allow us to exactly recreate an existing object.
    """

    batch = BatchTrial(experiment=experiment, ttl_seconds=ttl_seconds)
    batch._index = index
    batch._trial_type = trial_type
    batch._status = status
    batch._time_created = time_created
    batch._time_completed = time_completed
    batch._time_staged = time_staged
    batch._time_run_started = time_run_started
    batch._abandoned_reason = abandoned_reason
    batch._failed_reason = failed_reason
    batch._run_metadata = run_metadata or {}
    batch._stop_metadata = stop_metadata or {}
    batch._generator_run_structs = generator_run_structs
    batch._runner = runner
    batch._abandoned_arms_metadata = abandoned_arms_metadata
    batch._num_arms_created = num_arms_created
    batch._status_quo = status_quo
    batch._status_quo_weight_override = status_quo_weight_override
    batch.optimize_for_power = optimize_for_power
    batch._generation_step_index = generation_step_index
    batch._lifecycle_stage = lifecycle_stage
    batch._properties = properties
    batch._refresh_arms_by_name()  # Trigger cache build
    warn_on_kwargs(callable_with_kwargs=BatchTrial, **kwargs)
    return batch


def trial_from_json(
    experiment: core.experiment.Experiment,
    index: int,
    trial_type: Optional[str],
    status: TrialStatus,
    time_created: datetime,
    time_completed: Optional[datetime],
    time_staged: Optional[datetime],
    time_run_started: Optional[datetime],
    abandoned_reason: Optional[str],
    run_metadata: Optional[Dict[str, Any]],
    generator_run: GeneratorRun,
    runner: Optional[Runner],
    num_arms_created: int,
    # Allowing default values for backwards compatibility with
    # objects stored before these fields were added.
    failed_reason: Optional[str] = None,
    ttl_seconds: Optional[int] = None,
    generation_step_index: Optional[int] = None,
    properties: Optional[Dict[str, Any]] = None,
    stop_metadata: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> Trial:
    """Load Ax trial from JSON.

    Other classes don't need explicit deserializers, because we can just use
    their constructors (see decoder.py). However, the constructor for Trial
    does not allow us to exactly recreate an existing object.
    """

    trial = Trial(
        experiment=experiment, generator_run=generator_run, ttl_seconds=ttl_seconds
    )
    trial._index = index
    trial._trial_type = trial_type
    # Swap `DISPATCHED` for `RUNNING`, since `DISPATCHED` is deprecated and nearly
    # equivalent to `RUNNING`.
    trial._status = status if status != TrialStatus.DISPATCHED else TrialStatus.RUNNING
    trial._time_created = time_created
    trial._time_completed = time_completed
    trial._time_staged = time_staged
    trial._time_run_started = time_run_started
    trial._abandoned_reason = abandoned_reason
    trial._failed_reason = failed_reason
    trial._run_metadata = run_metadata or {}
    trial._stop_metadata = stop_metadata or {}
    trial._runner = runner
    trial._num_arms_created = num_arms_created
    trial._generation_step_index = generation_step_index
    trial._properties = properties or {}
    warn_on_kwargs(callable_with_kwargs=Trial, **kwargs)
    return trial


def transform_type_from_json(object_json: Dict[str, Any]) -> Type[Transform]:
    """Load the transform type from JSON."""
    index_in_registry = object_json.pop("index_in_registry")
    if index_in_registry not in REVERSE_TRANSFORM_REGISTRY:
        raise ValueError(f"Unknown transform '{object_json.pop('transform_type')}'")
    return REVERSE_TRANSFORM_REGISTRY[index_in_registry]


def input_transform_type_from_json(object_json: Dict[str, Any]) -> Type[InputTransform]:
    input_transform_type = object_json.pop("index")
    if input_transform_type not in REVERSE_INPUT_TRANSFORM_REGISTRY:
        raise ValueError(f"Unknown transform {input_transform_type}.")
    return REVERSE_INPUT_TRANSFORM_REGISTRY[input_transform_type]


def outcome_transform_type_from_json(
    object_json: Dict[str, Any]
) -> Type[OutcomeTransform]:
    outcome_transform_type = object_json.pop("index")
    if outcome_transform_type not in REVERSE_OUTCOME_TRANSFORM_REGISTRY:
        raise ValueError(f"Unknown transform {outcome_transform_type}.")
    return REVERSE_OUTCOME_TRANSFORM_REGISTRY[outcome_transform_type]


# pyre-fixme[3]: Return annotation cannot contain `Any`.
def class_from_json(json: Dict[str, Any]) -> Type[Any]:
    """Load any class registered in `CLASS_DECODER_REGISTRY` from JSON."""
    index_in_registry = json.pop("index")
    class_path = json.pop("class")
    for _class in CLASS_TO_REVERSE_REGISTRY:
        if class_path == f"{_class}":
            reverse_registry = CLASS_TO_REVERSE_REGISTRY[_class]
            if index_in_registry not in reverse_registry:
                raise ValueError(
                    f"Index '{index_in_registry}'"
                    " is not registered in the reverse registry."
                )
            return reverse_registry[index_in_registry]
    raise ValueError(
        f"{class_path} does not have a corresponding entry in "
        "CLASS_TO_REVERSE_REGISTRY."
    )


def tensor_from_json(json: Dict[str, Any]) -> torch.Tensor:
    try:
        device = (
            checked_cast(
                torch.device,
                torch_type_from_str(
                    identifier=json["device"]["value"], type_name="device"
                ),
            )
            if torch.cuda.is_available()
            else torch.device("cpu")
        )
        return torch.tensor(
            json["value"],
            dtype=checked_cast(
                torch.dtype,
                torch_type_from_str(
                    identifier=json["dtype"]["value"], type_name="dtype"
                ),
            ),
            device=device,
        )
    except KeyError as e:
        raise JSONDecodeError(
            f"Got KeyError {e} while attempting to construct a tensor from json. "
            f"Expected value, dtype, and device fields; got {json=}."
        )


def tensor_or_size_from_json(json: Dict[str, Any]) -> Union[torch.Tensor, torch.Size]:
    if json["__type"] == "Tensor":
        return tensor_from_json(json)
    elif json["__type"] == "torch_Size":
        return checked_cast(
            torch.Size,
            torch_type_from_str(identifier=json["value"], type_name="Size"),
        )
    else:
        raise JSONDecodeError(
            f"Expected json encoding of a torch.Tensor or torch.Size. Got {json=}"
        )


# pyre-fixme[3]: Return annotation cannot contain `Any`.
# pyre-fixme[2]: Parameter annotation cannot be `Any`.
def botorch_component_from_json(botorch_class: Any, json: Dict[str, Any]) -> Type[Any]:
    """Load any instance of `torch.nn.Module` or descendants registered in
    `CLASS_DECODER_REGISTRY` from state dict."""
    state_dict = json.pop("state_dict")
    if issubclass(botorch_class, ChainedInputTransform):
        return botorch_class(
            **{
                k: botorch_component_from_json(
                    botorch_class=REVERSE_INPUT_TRANSFORM_REGISTRY[v.pop("__type")],
                    json=v,
                )
                for k, v in state_dict.items()
            }
        )
    if issubclass(botorch_class, ChainedOutcomeTransform):
        return botorch_class(
            **{
                k: botorch_component_from_json(
                    botorch_class=REVERSE_OUTCOME_TRANSFORM_REGISTRY[v.pop("__type")],
                    json=v,
                )
                for k, v in state_dict.items()
            }
        )
    class_path = json.pop("class")
    init_args = inspect.signature(botorch_class).parameters
    required_args = {
        p for p, v in init_args.items() if v.default is inspect._empty and p != "kwargs"
    }
    allowable_args = set(init_args)
    received_args = set(state_dict)
    missing_args = required_args - received_args
    if missing_args:
        raise ValueError(
            f"Missing required initialization args {missing_args} for class "
            f"{class_path}. For gpytorch objects, this is likely because the "
            "object's `state_dict` method does not return the args required "
            "for initialization."
        )
    extra_args = received_args - allowable_args
    if extra_args:
        raise ValueError(
            f"Received unused args {extra_args} for class {class_path}. "
            "For gpytorch objects, this is likely because the object's "
            "`state_dict` method returns these extra args, which could "
            "indicate that the object's state will not be fully recreated "
            "by this serialization/deserialization method."
        )
    return botorch_class(
        **{
            k: (
                tensor_or_size_from_json(json=v)
                if isinstance(v, dict) and "__type" in v
                else v
            )
            for k, v in state_dict.items()
        }
    )


def pathlib_from_json(pathsegments: Union[str, Iterable[str]]) -> Path:
    if isinstance(pathsegments, str):
        return Path(pathsegments)

    return Path(*pathsegments)


def default_from_json(json: Dict[str, Any]) -> _DefaultType:
    if json != {}:
        raise JSONDecodeError(
            f"Expected empty json object for ``DEFAULT``, got {json=}"
        )
    return DEFAULT



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import plotly.graph_objects as go
from ax.modelbridge.cross_validation import cross_validate
from ax.modelbridge.registry import Models
from ax.plot.base import AxPlotConfig
from ax.plot.diagnostic import (
    interact_cross_validation,
    interact_cross_validation_plotly,
)
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment
from ax.utils.testing.mock import fast_botorch_optimize


class DiagnosticTest(TestCase):
    @fast_botorch_optimize
    def setUp(self) -> None:
        super().setUp()
        exp = get_branin_experiment(with_batch=True)
        exp.trials[0].run()
        self.model = Models.BOTORCH_MODULAR(
            # Model bridge kwargs
            experiment=exp,
            data=exp.fetch_data(),
        )

    def test_cross_validation(self) -> None:
        for autoset_axis_limits in [False, True]:
            cv = cross_validate(self.model)
            # Assert that each type of plot can be constructed successfully
            label_dict = {"branin": "BrAnIn"}
            plot = interact_cross_validation_plotly(
                cv, label_dict=label_dict, autoset_axis_limits=autoset_axis_limits
            )
            # pyre-ignore [16]
            x_range = plot.layout.updatemenus[0].buttons[0].args[1]["xaxis.range"]
            y_range = plot.layout.updatemenus[0].buttons[0].args[1]["yaxis.range"]
            if autoset_axis_limits:
                self.assertTrue((len(x_range) == 2) and (x_range[0] < x_range[1]))
                self.assertTrue((len(y_range) == 2) and (y_range[0] < y_range[1]))
            else:
                self.assertIsNone(x_range)
                self.assertIsNone(y_range)

            self.assertIsInstance(plot, go.Figure)
            plot = interact_cross_validation(
                cv, label_dict=label_dict, autoset_axis_limits=autoset_axis_limits
            )
            self.assertIsInstance(plot, AxPlotConfig)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import os
import pkgutil
from typing import List, Optional

from ax.plot.render import _js_requires, _load_css_resource as _load_plot_css_resource
from jinja2 import Environment, FunctionLoader

REPORT_MODULE_NAME = "ax.utils.report"


def _load_css_resource() -> str:

    resource = pkgutil.get_data(
        REPORT_MODULE_NAME, os.path.join("resources", "report.css")
    )
    assert resource is not None
    return resource.decode("utf8")


REPORT_ELEMENT_TEMPLATE = "simple_template.html"


def p_html(text: str) -> str:
    """Embed text in paragraph tag."""
    return "<p>{}</p>".format(text)


def h2_html(text: str) -> str:
    """Embed text in subheading tag."""
    return "<h2>{}</h2>".format(text)


def h3_html(text: str) -> str:
    """Embed text in subsubheading tag."""
    return "<h3>{}</h3>".format(text)


def list_item_html(text: str) -> str:
    """Embed text in list element tag."""
    return "<li>{}</li>".format(text)


def unordered_list_html(list_items: List[str]) -> str:
    """Embed list of html elements into an unordered list tag."""
    return "<ul>{}</ul>".format("".join(list_items))


def link_html(text: str, href: str) -> str:
    """Embed text and reference address into link tag."""
    return '<a href="{}">{}</a>'.format(href, text)


def table_cell_html(text: str, width: Optional[str] = None) -> str:
    """Embed text or an HTML element into table cell tag."""
    if width:
        return f"<td width={width}>{text}</td>"
    else:
        return f"<td>{text}</td>"


def table_heading_cell_html(text: str) -> str:
    """Embed text or an HTML element into table heading cell tag."""
    return "<th>{}</th>".format(text)


def table_row_html(table_cells: List[str]) -> str:
    """Embed list of HTML elements into table row tag."""
    return "<tr>{}</tr>".format("".join(table_cells))


def table_html(table_rows: List[str]) -> str:
    """Embed list of HTML elements into table tag."""
    return "<table>{}</table>".format("".join(table_rows))


def render_report_elements(
    experiment_name: str,
    html_elements: List[str],
    header: bool = True,
    offline: bool = False,
    notebook_env: bool = False,
) -> str:
    """Generate Ax HTML report for a given experiment from HTML elements.

    Uses Jinja2 for template. Injects Plotly JS for graph rendering.

    Example:

    ::

        html_elements = [
            h2_html("Subsection with plot"),
            p_html("This is an example paragraph."),
            plot_html(plot_fitted(gp_model, 'perf_metric')),
            h2_html("Subsection with table"),
            pandas_html(data.df),
        ]
        html = render_report_elements('My experiment', html_elements)

    Args:
        experiment_name: the name of the experiment to use for title.
        html_elements: list of HTML strings to render in report
            body.
        header: if True, render experiment title as a header.
            Meant to be used for standalone reports (e.g. via email), as opposed
            to served on the front-end.
        offline: if True, entire Plotly library is bundled
            with report.
        notebook_env: if True, caps the report width to 700px for viewing in a
            notebook environment.

    Returns:
        str: HTML string.

    """
    # combine CSS for report and plots
    css = _load_css_resource() + _load_plot_css_resource()
    return (
        _get_jinja_environment()
        .get_template(REPORT_ELEMENT_TEMPLATE)
        .render(
            experiment_name=experiment_name,
            css=css,
            js_requires=_js_requires(),
            html_elements=html_elements,
            headfoot=header,
            notebook_env=notebook_env,
        )
    )


def _load_html_template(name: str) -> str:

    resource = pkgutil.get_data(REPORT_MODULE_NAME, os.path.join("resources", name))
    assert resource is not None
    return resource.decode("utf8")


def _get_jinja_environment() -> Environment:
    return Environment(loader=FunctionLoader(_load_html_template))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import tempfile
from unittest.mock import patch

import numpy as np
from ax.benchmark.benchmark import (
    _create_benchmark_experiment,
    benchmark_multiple_problems_methods,
    benchmark_one_method_problem,
    benchmark_replication,
    make_ground_truth_metrics,
    make_ground_truth_optimization_config,
)
from ax.benchmark.benchmark_method import (
    BenchmarkMethod,
    get_benchmark_scheduler_options,
)
from ax.benchmark.benchmark_problem import SingleObjectiveBenchmarkProblem
from ax.benchmark.benchmark_result import BenchmarkResult
from ax.benchmark.methods.modular_botorch import get_sobol_botorch_modular_acquisition
from ax.benchmark.metrics.base import GroundTruthMetricMixin
from ax.benchmark.metrics.benchmark import BenchmarkMetric, GroundTruthBenchmarkMetric
from ax.benchmark.problems.registry import get_problem
from ax.modelbridge.generation_strategy import GenerationNode, GenerationStrategy
from ax.modelbridge.model_spec import ModelSpec
from ax.modelbridge.registry import Models
from ax.service.utils.scheduler_options import SchedulerOptions
from ax.storage.json_store.load import load_experiment
from ax.storage.json_store.save import save_experiment
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast, not_none
from ax.utils.testing.benchmark_stubs import (
    get_constrained_multi_objective_benchmark_problem,
    get_moo_surrogate,
    get_multi_objective_benchmark_problem,
    get_single_objective_benchmark_problem,
    get_sobol_benchmark_method,
    get_soo_surrogate,
)
from ax.utils.testing.core_stubs import get_experiment
from ax.utils.testing.mock import fast_botorch_optimize
from botorch.acquisition.logei import qLogNoisyExpectedImprovement
from botorch.acquisition.multi_objective.monte_carlo import (
    qNoisyExpectedHypervolumeImprovement,
)
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.gp_regression import SingleTaskGP
from botorch.optim.optimize import optimize_acqf
from botorch.test_functions.synthetic import Branin


class TestBenchmark(TestCase):
    @fast_botorch_optimize
    def test_batch(self) -> None:
        batch_size = 5

        problem = get_problem("ackley4", num_trials=2)
        batch_options = get_benchmark_scheduler_options(batch_size=batch_size)
        for sequential in [False, True]:
            with self.subTest(sequential=sequential):
                batch_method_joint = get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    scheduler_options=batch_options,
                    distribute_replications=False,
                    model_gen_kwargs={
                        "model_gen_options": {
                            "optimizer_kwargs": {"sequential": sequential}
                        }
                    },
                    num_sobol_trials=1,
                )
                # this is generating more calls to optimize_acqf than expected
                with patch(
                    "ax.models.torch.botorch_modular.acquisition.optimize_acqf",
                    wraps=optimize_acqf,
                ) as mock_optimize_acqf:
                    benchmark_one_method_problem(
                        problem=problem, method=batch_method_joint, seeds=[0]
                    )
                mock_optimize_acqf.assert_called_once()
                self.assertEqual(
                    mock_optimize_acqf.call_args.kwargs["sequential"], sequential
                )
                self.assertEqual(mock_optimize_acqf.call_args.kwargs["q"], batch_size)

    def test_storage(self) -> None:
        problem = get_single_objective_benchmark_problem()
        res = benchmark_replication(
            problem=problem, method=get_sobol_benchmark_method(), seed=0
        )
        # Experiment is not in storage yet
        self.assertTrue(res.experiment is not None)
        self.assertEqual(res.experiment_storage_id, None)
        experiment = res.experiment

        # test saving to temporary file
        with tempfile.NamedTemporaryFile(mode="w", delete=True, suffix=".json") as f:
            save_experiment(not_none(res.experiment), f.name)
            res.experiment_storage_id = f.name
            res.experiment = None
            self.assertIsNone(res.experiment)
            self.assertEqual(res.experiment_storage_id, f.name)

            # load it back
            experiment = load_experiment(f.name)
            self.assertEqual(experiment, experiment)

    def test_make_ground_truth_metrics(self) -> None:
        problem = get_single_objective_benchmark_problem(observe_noise_sd=False)
        metric = problem.optimization_config.objective.metric

        # basic setup
        gt_metrics = make_ground_truth_metrics(problem=problem)
        self.assertEqual(len(gt_metrics), 1)
        gt_metric = checked_cast(GroundTruthBenchmarkMetric, gt_metrics[metric.name])
        self.assertIs(gt_metric.original_metric, metric)

        # add a tracking metric
        tracking_metric = BenchmarkMetric(name="test_track", lower_is_better=True)
        problem.tracking_metrics = [tracking_metric]
        gt_metrics = make_ground_truth_metrics(problem=problem)
        self.assertEqual(len(gt_metrics), 2)
        gt_tracking_metric = checked_cast(
            GroundTruthBenchmarkMetric, gt_metrics["test_track"]
        )
        self.assertIs(gt_tracking_metric.original_metric, tracking_metric)

        # set include_tracking_metrics=False
        gt_metrics = make_ground_truth_metrics(
            problem=problem, include_tracking_metrics=False
        )
        self.assertEqual(len(gt_metrics), 1)

        # error out if the problem does not have ground truth
        problem.has_ground_truth = False
        with self.assertRaisesRegex(ValueError, "do not have a ground truth"):
            make_ground_truth_metrics(problem=problem)

    def test_make_ground_truth_optimization_config(self) -> None:
        problem = get_single_objective_benchmark_problem(observe_noise_sd=False)
        metric = problem.optimization_config.objective.metric
        experiment = _create_benchmark_experiment(
            problem=problem, method_name="test_method"
        )

        # A vanilla experiment w/o ground truth metrics attached should error
        with self.assertRaisesRegex(
            ValueError, f"Ground truth metric for metric {metric.name} not found!"
        ):
            make_ground_truth_optimization_config(experiment)

        # Add the ground truth metric and check basic behavior
        gt_metric = make_ground_truth_metrics(problem)[metric.name]
        experiment.add_tracking_metric(gt_metric)
        gt_opt_cfg = make_ground_truth_optimization_config(experiment)
        self.assertIs(gt_opt_cfg.objective.metric, gt_metric)

        # Test behavior with MOO problem and outcome constraints
        problem = get_constrained_multi_objective_benchmark_problem(
            observe_noise_sd=False
        )
        experiment = _create_benchmark_experiment(
            problem=problem, method_name="test_method"
        )
        gt_metrics = make_ground_truth_metrics(problem)
        for metric in problem.optimization_config.objective.metrics:
            experiment.add_tracking_metric(gt_metrics[metric.name])
        gt_opt_cfg = make_ground_truth_optimization_config(experiment)

        for metric in gt_opt_cfg.objective.metrics:
            gt_name = metric.name
            metric = checked_cast(GroundTruthMetricMixin, metric)
            self.assertIs(metric, gt_metrics[metric.get_original_name(gt_name)])

        for metric in gt_opt_cfg.outcome_constraints:
            gt_name = metric.metric.name
            metric = checked_cast(GroundTruthMetricMixin, metric.metric)
            self.assertIs(metric, gt_metrics[metric.get_original_name(gt_name)])

    def test_benchmark_result_invalid_inputs(self) -> None:
        """
        Test that a BenchmarkResult cannot be specified with both an `experiment`
        and an `experiment_storage_id`.
        """
        with self.assertRaisesRegex(ValueError, "Cannot specify both an `experiment` "):
            BenchmarkResult(
                name="name",
                seed=0,
                optimization_trace=np.array([]),
                score_trace=np.array([]),
                fit_time=0.0,
                gen_time=0.0,
                experiment=get_experiment(),
                experiment_storage_id="experiment_storage_id",
            )

        with self.assertRaisesRegex(
            ValueError, "Must provide an `experiment` or `experiment_storage_id`"
        ):
            BenchmarkResult(
                name="name",
                seed=0,
                optimization_trace=np.array([]),
                score_trace=np.array([]),
                fit_time=0.0,
                gen_time=0.0,
            )

    def test_create_benchmark_experiment(self) -> None:

        with self.subTest("noiseless"):
            problem = get_single_objective_benchmark_problem(observe_noise_sd=False)
            experiment = _create_benchmark_experiment(
                problem=problem, method_name="test_method"
            )
            self.assertTrue("|test_method_" in experiment.name)
            self.assertFalse("_observed_noise" in experiment.name)
            self.assertEqual(experiment.search_space, problem.search_space)
            self.assertEqual(
                experiment.optimization_config, problem.optimization_config
            )
            self.assertEqual(len(experiment.tracking_metrics), 0)
            self.assertEqual(experiment.runner, problem.runner)

        with self.subTest("noisy, unobserved noise std"):
            problem = get_single_objective_benchmark_problem(
                observe_noise_sd=False, test_problem_kwargs={"noise_std": 0.1}
            )
            experiment = _create_benchmark_experiment(
                problem=problem, method_name="test_method"
            )
            self.assertTrue("|test_method_" in experiment.name)
            self.assertFalse("_observed_noise" in experiment.name)
            self.assertEqual(experiment.search_space, problem.search_space)
            self.assertEqual(
                experiment.optimization_config, problem.optimization_config
            )
            self.assertEqual(len(experiment.tracking_metrics), 1)
            gt_metric = checked_cast(
                GroundTruthBenchmarkMetric, experiment.tracking_metrics[0]
            )
            self.assertIs(
                gt_metric.original_metric,
                problem.optimization_config.objective.metric,
            )
            self.assertEqual(experiment.runner, problem.runner)

        with self.subTest("noisy, observed noise std"):
            problem = get_single_objective_benchmark_problem(
                observe_noise_sd=True, test_problem_kwargs={"noise_std": 0.1}
            )
            experiment = _create_benchmark_experiment(
                problem=problem, method_name="test_method"
            )
            self.assertTrue("|test_method_" in experiment.name)
            self.assertTrue("_observed_noise" in experiment.name)
            self.assertEqual(experiment.search_space, problem.search_space)
            self.assertEqual(
                experiment.optimization_config, problem.optimization_config
            )
            self.assertEqual(len(experiment.tracking_metrics), 1)
            gt_metric = checked_cast(
                GroundTruthBenchmarkMetric, experiment.tracking_metrics[0]
            )
            self.assertIs(
                gt_metric.original_metric,
                problem.optimization_config.objective.metric,
            )
            self.assertEqual(experiment.runner, problem.runner)

    def test_replication_sobol_synthetic(self) -> None:
        method = get_sobol_benchmark_method()
        problem = get_single_objective_benchmark_problem()
        res = benchmark_replication(problem=problem, method=method, seed=0)

        self.assertEqual(
            min(problem.num_trials, not_none(method.scheduler_options.total_trials)),
            len(not_none(res.experiment).trials),
        )

        self.assertTrue(np.isfinite(res.score_trace).all())
        self.assertTrue(np.all(res.score_trace <= 100))

    def test_replication_sobol_surrogate(self) -> None:
        method = get_sobol_benchmark_method()

        # This is kind of a weird setup - these are "surrogates" that use a Branin
        # synthetic function. The idea here is to test the machinery around the
        # surrogate benchmarks without having to actually load a surrogate model
        # of potentially non-neglible size.
        for name, problem in [
            ("soo", get_soo_surrogate()),
            ("moo", get_moo_surrogate()),
        ]:
            with self.subTest(name, problem=problem):
                surrogate, datasets = not_none(problem.get_surrogate_and_datasets)()
                res = benchmark_replication(problem=problem, method=method, seed=0)

                self.assertEqual(
                    min(
                        problem.num_trials,
                        not_none(method.scheduler_options.total_trials),
                    ),
                    len(not_none(res.experiment).trials),
                )

                self.assertTrue(np.isfinite(res.score_trace).all())
                self.assertTrue(np.all(res.score_trace <= 100))

    @fast_botorch_optimize
    def test_replication_mbm(self) -> None:
        for method, problem, expected_name in [
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    distribute_replications=True,
                ),
                get_problem("constrained_gramacy_observed_noise", num_trials=6),
                "MBM::SingleTaskGP_qLogNEI",
            ),
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    scheduler_options=get_benchmark_scheduler_options(),
                    distribute_replications=False,
                ),
                get_single_objective_benchmark_problem(
                    observe_noise_sd=True, num_trials=6
                ),
                "MBM::SingleTaskGP_qLogNEI",
            ),
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    distribute_replications=False,
                ),
                get_single_objective_benchmark_problem(
                    observe_noise_sd=True, num_trials=6
                ),
                "MBM::SingleTaskGP_qLogNEI",
            ),
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qNoisyExpectedHypervolumeImprovement,
                    distribute_replications=False,
                ),
                get_multi_objective_benchmark_problem(
                    observe_noise_sd=True, num_trials=6
                ),
                "MBM::SingleTaskGP_qNEHVI",
            ),
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SaasFullyBayesianSingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    distribute_replications=False,
                ),
                get_multi_objective_benchmark_problem(num_trials=6),
                "MBM::SAAS_qLogNEI",
            ),
        ]:
            with self.subTest(method=method, problem=problem):
                res = benchmark_replication(problem=problem, method=method, seed=0)
                self.assertEqual(
                    problem.num_trials,
                    len(not_none(res.experiment).trials),
                )
                self.assertTrue(np.all(res.score_trace <= 100))
                self.assertEqual(method.name, method.generation_strategy.name)
                self.assertEqual(method.name, expected_name)

    def test_replication_moo_sobol(self) -> None:
        problem = get_multi_objective_benchmark_problem()

        res = benchmark_replication(
            problem=problem, method=get_sobol_benchmark_method(), seed=0
        )

        self.assertEqual(
            problem.num_trials,
            len(not_none(res.experiment).trials),
        )
        self.assertEqual(
            problem.num_trials * 2,
            len(not_none(res.experiment).fetch_data().df),
        )

        self.assertTrue(np.all(res.score_trace <= 100))

    def test_benchmark_one_method_problem(self) -> None:
        problem = get_single_objective_benchmark_problem()
        agg = benchmark_one_method_problem(
            problem=problem,
            method=get_sobol_benchmark_method(),
            seeds=(0, 1),
        )

        self.assertEqual(len(agg.results), 2)
        self.assertTrue(
            all(
                len(not_none(result.experiment).trials) == problem.num_trials
                for result in agg.results
            ),
            "All experiments must have 4 trials",
        )

        for col in ["mean", "P25", "P50", "P75"]:
            self.assertTrue((agg.score_trace[col] <= 100).all())

    @fast_botorch_optimize
    def test_benchmark_multiple_problems_methods(self) -> None:
        aggs = benchmark_multiple_problems_methods(
            problems=[get_single_objective_benchmark_problem(num_trials=6)],
            methods=[
                get_sobol_benchmark_method(),
                get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    distribute_replications=False,
                ),
            ],
            seeds=(0, 1),
        )

        self.assertEqual(len(aggs), 2)
        for agg in aggs:
            for col in ["mean", "P25", "P50", "P75"]:
                self.assertTrue((agg.score_trace[col] <= 100).all())

    def test_timeout(self) -> None:
        problem = SingleObjectiveBenchmarkProblem.from_botorch_synthetic(
            test_problem_class=Branin,
            test_problem_kwargs={},
            lower_is_better=True,
            num_trials=1000,  # Unachievable num_trials
        )

        generation_strategy = get_sobol_botorch_modular_acquisition(
            model_cls=SingleTaskGP,
            acquisition_cls=qLogNoisyExpectedImprovement,
            distribute_replications=False,
            num_sobol_trials=1000,  # Ensures we don't use BO
        ).generation_strategy

        method = BenchmarkMethod(
            name=generation_strategy.name,
            generation_strategy=generation_strategy,
            scheduler_options=SchedulerOptions(
                max_pending_trials=1,
                init_seconds_between_polls=0,
                min_seconds_before_poll=0,
                timeout_hours=0.0001,  # Strict timeout of 0.36 seconds
            ),
        )

        # Each replication will have a different number of trials
        result = benchmark_one_method_problem(
            problem=problem, method=method, seeds=(0, 1)
        )

        # Test the traces get composited correctly. The AggregatedResult's traces
        # should be the length of the shortest trace in the BenchmarkResults
        min_num_trials = min(len(res.optimization_trace) for res in result.results)
        self.assertEqual(len(result.optimization_trace), min_num_trials)
        self.assertEqual(len(result.score_trace), min_num_trials)

    def test_replication_with_generation_node(self) -> None:
        method = BenchmarkMethod(
            name="Sobol Generation Node",
            generation_strategy=GenerationStrategy(
                nodes=[
                    GenerationNode(
                        node_name="Sobol",
                        model_specs=[
                            ModelSpec(Models.SOBOL, model_kwargs={"deduplicate": True})
                        ],
                    )
                ]
            ),
            scheduler_options=SchedulerOptions(),
        )
        problem = get_single_objective_benchmark_problem()
        res = benchmark_replication(problem=problem, method=method, seed=0)

        self.assertEqual(problem.num_trials, len(not_none(res.experiment).trials))
        self.assertTrue(np.isnan(res.score_trace).all())
